using System;
using System.Collections.Generic;
using System.Text;
using WinFormsGraphicsDevice;
using SharpMedia.XNA;
using Microsoft.Xna.Framework;
using System.Runtime.InteropServices;
using System.Security;
using System.Threading;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using PhysicsEditor;
using GameEngine.Helpers;
using GameEngine.ScreenSystem;
using Editor.Tools;
using Microsoft.Xna.Framework.Input;
using Editor.Data;
using GameEngine;
using GameEngine.Xml;
using System.Linq;
using GameEngine.Cameras;
using GamePrototype;

namespace Editor
{
    public class ToolChangeEventArgs : EventArgs
    {
        private string oldValue, newValue;
        public string OldTool { get { return this.oldValue; } }
        public string NewTool { get { return this.newValue; } }

        public ToolChangeEventArgs(string oldValue, string newValue)
        {
            this.oldValue = oldValue;
            this.newValue = newValue;
        }
    }

    public class GridVisibilityChangedEventArgs : EventArgs
    {
        private string whichGrid;
        public string WhichGrid { get { return this.whichGrid; } }
        private bool newValue;
        public bool NewValue { get { return this.newValue; } }

        public GridVisibilityChangedEventArgs(string whichGrid, bool newValue)
        {
            this.whichGrid = whichGrid;
            this.newValue = newValue;
        }
    }

    public delegate void ToolChangedHandler(object sender, ToolChangeEventArgs e);
    public delegate void GridVisibilityChangedHandler(object sender, GridVisibilityChangedEventArgs e);

    public class EditorControl : GraphicsDeviceControl
    {
        bool dirty = true;
        public bool Dirty { get { return dirty; } set { dirty = value; } }
        private bool active = true;

        public bool Active
        {
            get { return active; }
            set { active = value; }
        }

        protected long ticksAtLastRender = XNAUtil.CurrentTickCount;
        public ContentManager Content, GameContent;
        public Effect primitivesEffect;
        public SpriteFont screenFont;
        SpriteBatch spriteBatch;
        PolygonDrawer polyDrawer;
        public OrthoOrbitCamera camera;
        public DrawInfo drawInfo;
        public Grid xzGrid;
        public bool XZGridVisible { get { return xzGrid.Visible; } set { xzGrid.Visible = value; } }
        public Grid xyGrid;
        public bool XYGridVisible { get { return xyGrid.Visible; } set { xyGrid.Visible = value; } }


        private GeometryDrawer geometryDrawer;
        private DeferredRenderer renderer;

        public ForestHandler forest;

        private Texture2D cursorTexture;

//        public PhysicsPolygonHandler polygonHandler;
        public ToolHandler ToolHandler;
        public HelperDrawer helperDrawer;
        public TilesetDebug tilesetDebugDrawer;


        private string toolName = "SelectionTool";
        public string ToolName { get { return toolName; } }
        public event ToolChangedHandler ToolChanged;
        public event GridVisibilityChangedHandler GridVisibilityChanged;
        public event SelectionChangedHandler SelectionChanged;
        public event EventHandler Initialized;

        

        public EditorControl()
        {

        }

        void graphicsDeviceService_DeviceReset(object sender, EventArgs e)
        {
            Util.NullDispose(ref renderer);
            renderer = new DeferredRenderer(GraphicsDevice.PresentationParameters.BackBufferWidth, GraphicsDevice.PresentationParameters.BackBufferHeight);
            renderer.LoadContent(GraphicsDevice, GameContent);
        }


        protected virtual void OnToolChanged(ToolChangeEventArgs e)
        {
            if (ToolChanged != null)
                ToolChanged(this, e);
        }
        
        public void SetTool(string newTool)
        {
            string oldTool = ToolHandler.ActiveTool.Name;
            ToolHandler.SelectTool(newTool);
            OnToolChanged(new ToolChangeEventArgs(oldTool, newTool));
        }

        protected virtual void OnGridVisibilityChanged(GridVisibilityChangedEventArgs e)
        {
            if (GridVisibilityChanged != null)
                GridVisibilityChanged(this, e);
        }

        public void ToggleGridVisibility(string whichGrid)
        {
            switch (whichGrid)
            {
                case "xy":
                    xyGrid.Visible = !xyGrid.Visible;
                    OnGridVisibilityChanged(new GridVisibilityChangedEventArgs("xy", xyGrid.Visible));
                    break;
                case "xz":
                    xzGrid.Visible = !xzGrid.Visible;
                    OnGridVisibilityChanged(new GridVisibilityChangedEventArgs("xz", xzGrid.Visible));
                    break;
                default:
                    break;
            }
        }

        public enum GridChange
	    {
            Larger,
            Smaller	         
	    }

        public void GridResize(string whichGrid, GridChange change)
        {
            switch (whichGrid)
            {
                case "xz":
                    switch (change)
                    {
                        case GridChange.Larger:
                            xzGrid.ResizeLarger();
                            break;
                        case GridChange.Smaller:
                        default:
                            xzGrid.ResizeSmaller();
                            break;
                    }
                    break;
                case "xy":
                default:
                    switch (change)
                    {
                        case GridChange.Larger:
                            xyGrid.ResizeLarger();
                            break;
                        case GridChange.Smaller:
                        default:
                            xyGrid.ResizeSmaller();
                            break;
                    }
                    break;
            }
        }

        public void SetGridVisibility(string whichGrid, bool newValue)
        {
            switch (whichGrid)
            {
                case "xy":
                    xyGrid.Visible = newValue;
                    OnGridVisibilityChanged(new GridVisibilityChangedEventArgs("xy", newValue));
                    break;
                case "xz":
                    xzGrid.Visible = newValue;
                    OnGridVisibilityChanged(new GridVisibilityChangedEventArgs("xz", newValue));
                    break;
                default:
                    break;
            }
        }

        public void OnSelectionChanged(SelectionChangedEventArgs e)
        {
            if (e.SelectedNodes.Count > 0 && e.SelectedNodes[0] is ObjectNode)
                camera.Target = ((ObjectNode)e.SelectedNodes[0]).InitialPosition;
            if(SelectionChanged != null)
                SelectionChanged(this, e);
        }

        void SelectionHandler_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            OnSelectionChanged(e);
        }

        public void CreateGrids(int gridSize, bool showXZ)
        {
            xzGrid = new Grid(gridSize, gridSize, 10f, new Color(0.5f,.5f,.5f,.5f));
            xyGrid = new Grid(gridSize, gridSize, 10f, Color.Orange);
            xyGrid.Visible = false;
            xzGrid.Visible = showXZ;
            xyGrid.Rotation = Quaternion.CreateFromAxisAngle(new Vector3(1, 0, 0), MathHelper.PiOver2);
            xzGrid.Initialize(this.graphicsDeviceService.GraphicsDevice, this.primitivesEffect);
            xyGrid.Initialize(this.graphicsDeviceService.GraphicsDevice, this.primitivesEffect);
        }

        protected override void Initialize()
        {
            this.graphicsDeviceService.DeviceReset += new EventHandler(graphicsDeviceService_DeviceReset);
            renderer = new DeferredRenderer(GraphicsDevice.PresentationParameters.BackBufferWidth, GraphicsDevice.PresentationParameters.BackBufferWidth);

            this.MouseWheel += new System.Windows.Forms.MouseEventHandler(EditorControl_MouseWheel);
            this.MouseMove += new System.Windows.Forms.MouseEventHandler(editorControl1_MouseMove);
            this.MouseEnter += new EventHandler(editorControl1_MouseEnter);
            this.MouseLeave += new EventHandler(editorControl1_MouseLeave);

            Content = new ContentManager(Services);
            Content.RootDirectory = "Content";
            drawInfo = new DrawInfo();

            renderer.LoadContent(this.GraphicsDevice, GameContent);

            CreateGrids(160, true);

            spriteBatch = new SpriteBatch(GraphicsDevice);

            //form.ContentLoad();

            primitivesEffect = Content.Load<Effect>("primitives");
            cursorTexture = Content.Load<Texture2D>("Hud/cursor");
            screenFont = Content.Load<SpriteFont>("Fonts/Arial");
            geometryDrawer = new GeometryDrawer();
            
            geometryDrawer.Load(GameContent, graphicsDeviceService.GraphicsDevice);
            
            camera = new OrthoOrbitCamera(MathHelper.PiOver4, new Vector3(0, 1250, 1250), Vector3.Zero, Vector3.Up, GraphicsDevice.Viewport, 1f, 100000f, CameraMode.Perspective);

            ToolHandler = new ToolHandler(camera);
            ToolHandler.LoadContent(Content);
            ToolHandler.GetTool<SelectionTool>().SetObjectCollections(selectableModels, polygons);
            ToolHandler.GetTool<PolyModifierTool>().SetObjectCollections(polygons);
            ToolHandler.Selection.SelectionChanged += new SelectionChangedHandler(SelectionHandler_SelectionChanged);

            helperDrawer = new HelperDrawer(primitivesEffect, GraphicsDevice);

            GameEngine.Helpers.Cursor.Initialize(cursorTexture, xyGrid, xzGrid, camera);

            xzGrid.Initialize(GraphicsDevice, primitivesEffect);
            xyGrid.Initialize(GraphicsDevice, primitivesEffect);
            
            tilesetDebugDrawer = new TilesetDebug(primitivesEffect, GraphicsDevice);

            drawInfo.GraphicsDevice = GraphicsDevice;
//            polygonHandler = new PhysicsPolygonHandler();
            controls.Add(this);
            this.Disposed += new EventHandler(EditorControl_Disposed);
            if (Initialized != null)
            {
                Initialized(this, null);
            }

            polyDrawer = new PolygonDrawer(primitivesEffect, GraphicsDevice);

            selectableModels.Clear();
            polygons.Clear();
            updateRoots.Clear();
            forest.TraverseBreadthFirst(UpdateUpdateables);
        }
        
        List<INodeUpdateable> updateRoots = new List<INodeUpdateable>();
        Queue<INodeRenderable> renderQueue = new Queue<INodeRenderable>();
        List<ObjectNode> selectableModels = new List<ObjectNode>();
        List<PolygonNode> polygons = new List<PolygonNode>();


        public void forest_NodeRemoved(GameEngine.ForestHandler forest, string path, GameEngine.Xml.SceneNode node)
        {
            updateRoots.Clear();
            selectableModels.Clear();
            polygons.Clear();
            forest.TraverseBreadthFirst(UpdateUpdateables);
            if (node is IPointLight)
                drawInfo.PointLights.Remove(node as IPointLight);

        }

        public void forest_NodeAdded(GameEngine.ForestHandler forest, string path, GameEngine.Xml.SceneNode node)
        {
            updateRoots.Clear();
            selectableModels.Clear();
            polygons.Clear();
            this.forest = forest;
            forest.TraverseBreadthFirst(UpdateUpdateables);
            if (node is IPointLight)
                drawInfo.PointLights.Add(node as IPointLight);
        }

        private void UpdateUpdateables(ForestHandler sender, string path, GameEngine.Xml.SceneNode node)
        {
            /*INodeUpdateable updatableNode = node as INodeUpdateable;
            if (updatableNode != null)
            {
                updatableNode.UpdateNodes.Clear();
                if (!(node.Parent is INodeUpdateable))
                {
                    updateRoots.Add(updatableNode);
                }
                else
                {
                    ((INodeUpdateable)node.Parent).UpdateNodes.Add(updatableNode);
                }
            }*/
            ObjectNode modelNode = node as ObjectNode;
            if (modelNode != null)
            {
                selectableModels.Add(modelNode);
            }
            PolygonNode polyNode = node as PolygonNode;
            if(polyNode != null)
            {
                polygons.Add(polyNode);
            }
        }

        void EditorControl_Disposed(object sender, EventArgs e)
        {
            controls.Remove(this);
        }

    

        protected override void OnKeyUp(System.Windows.Forms.KeyEventArgs e)
        {
            e.Handled = true;
            base.OnKeyUp(e);
        }

        void EditorControl_MouseWheel(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            scroll = e.Delta;
        }


        protected override void Draw(GameTime gt)
        {
            GraphicsDevice.Clear(Color.Black);
            //GraphicsDevice.RenderState.AlphaBlendEnable = false;

            drawInfo.PointLights.Clear();
            drawInfo.UpdateOBBs = false;

            Viewport vp = drawInfo.GraphicsDevice.Viewport;

            renderer.BeginRender(drawInfo);
            drawInfo.GraphicsDevice.Viewport = vp;
            forest.Roots.DrawRecursive(gt, drawInfo);

            if (drawInfo.PointLights.Count == 0)
                renderer.Ambient = 1f;
            else
                renderer.Ambient = 0.5f;


            geometryDrawer.Draw(drawInfo);

            renderer.EndRender(null);
            drawInfo.GraphicsDevice.Viewport = vp;
            
            GraphicsDevice.RenderState.AlphaBlendEnable = true;
            GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            GraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;
            xyGrid.Draw(drawInfo);
            xzGrid.Draw(drawInfo);
            GraphicsDevice.RenderState.AlphaBlendEnable = false;
            //spriteBatch.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            //spriteBatch.Draw(renderer.colorRT.GetTexture(), new Rectangle(0,0,GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height), Color.White);
            //spriteBatch.End();


            //while (renderQueue.Count > 0)
            //    renderQueue.Dequeue().Draw(gt, drawInfo);
            GraphicsDevice.RenderState.DepthBufferEnable = false;

            for (int i = 0; i < selectableModels.Count; i++)
            {
                if (ToolHandler.Selection.Contains(selectableModels[i] as ObjectNode))
                {
                    helperDrawer.DrawNodeBoundingBoxes(selectableModels[i], drawInfo);
                }
            }
            if (ToolHandler.Selection.Count > 1)
            {
                switch (ToolHandler.Selection.SelectionMode)
                {
                    case SelectionMode.Select2D:
                        break;
                    case SelectionMode.Select3D:
                        helperDrawer.DrawOrientedBoundingBox(Matrix.CreateFromQuaternion(ToolHandler.Selection.SelectionRotation), ToolHandler.Selection.SelectionExtent, ToolHandler.Selection.SelectionCenter, drawInfo, Color.CornflowerBlue);
                        break;
                    case SelectionMode.SelectPoints:
                        helperDrawer.DrawPoint(ToolHandler.Selection.SelectionCenter, drawInfo, Color.Red);
                        break;
                    default:
                        break;
                }
            }

            // If we're working with points, we want to see them
            if (ToolHandler.Selection.SelectionMode == SelectionMode.SelectPoints)
            {
                PolyModifierTool polyModTool = ToolHandler.GetTool<PolyModifierTool>();

                PolygonNode workPolyNode = polyModTool.ClosestPoint.Container;
                int closestIndex = polyModTool.ClosestPoint.Index;

                for (int i = 0; i < polygons.Count; i++)
                {
                    for (int j = 0; j < polygons[i].Points.Count; j++)
                    {
                        Color color = Color.Green;
                        if (workPolyNode == polygons[i] && j == closestIndex)
                        {
                                color = Color.Red;
                        }
                        else
                        {
                            if (ToolHandler.Selection.Contains(polygons[i], j))
                            {
                                color = Color.White;
                            }
                        }

                        helperDrawer.DrawPoint(new Vector3(polygons[i].Points[j].X, 0, polygons[i].Points[j].Y) + polygons[i].WorldPosition, drawInfo, color);
                    }
                }
            }

//            polygonHandler.Draw(drawInfo);
            ToolHandler.ActiveTool.Draw(drawInfo);


            ObjectNode[] selection = ToolHandler.Selection.Get3DSelection();

            polyDrawer.DrawPolys(forest, drawInfo, selection.Where(x => x is PolygonNode).Cast<PolygonNode>() );

            foreach (var node in forest.Roots)
            {
                TileInfoNode tin = node as TileInfoNode;
                if (tin != null)
                {
                    TilesetNode tn = tin.Tag as TilesetNode;
                    
                    if(tn == null) 
                        break;

                    tilesetDebugDrawer.DrawMarkers(tn.TileSize, tn.ExitSize, tin.TileType, drawInfo);
                }
            }
            


            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            GameEngine.Helpers.Cursor.Draw2D(spriteBatch);

            


            ToolHandler.Draw2D(spriteBatch, drawInfo);
            spriteBatch.End();

        }

        public void Update(GameTime gt)
        {
            forest.Roots.UpdateRecursive(gt, drawInfo);
            
            // get the point in which the grid intersects with the camera.
            Vector2? xyPos = GameEngine.Helpers.Cursor.GetXYPlanePosition(new Ray(camera.Position, Vector3.Normalize(camera.Target - camera.Position)), true);
            Vector2? xzPos = GameEngine.Helpers.Cursor.GetXZPlanePosition(new Ray(camera.Position, Vector3.Normalize(camera.Target - camera.Position)), true);
            /*
            if (xyPos != null)
                xyGrid.Position = new Vector3(xyPos.Value, 0);

            if (xzPos != null)
                xzGrid.Position = new Vector3(xzPos.Value.X, 0, xzPos.Value.Y);
            */


            xyGrid.Update();
            xzGrid.Update();
            //cursor.Update();
            //Cursor.Position -= this.l

            // Camera updates

            camera.Viewport = GraphicsDevice.Viewport;
            camera.UpdateCamera(gt);
            drawInfo.Camera = camera;

            
            ToolHandler.ActiveTool.HandleInput(gt);
            ToolHandler.ActiveTool.Update(gt);
            if (ToolHandler.ActiveTool.Name != toolName)
            {
                if(ToolChanged != null) 
                    ToolChanged(this, new ToolChangeEventArgs(toolName, ToolHandler.ActiveTool.Name));
                toolName = ToolHandler.ActiveTool.Name;
            }


            if (InputState.KeyPressed(Keys.Add))
            {
                xyGrid.ResizeLarger();                 
            }

            if (InputState.KeyPressed(Keys.Subtract))
            {
                xyGrid.ResizeSmaller();
            }

            if (InputState.KeyPressed(Keys.PageUp))
            {
                xzGrid.ResizeLarger();
            }

            if (InputState.KeyPressed(Keys.PageDown))
            {
                xzGrid.ResizeSmaller();
            }

            if (InputState.KeyPressed(Keys.F1))
            {
                xyGrid.Visible = !xyGrid.Visible;
                if( GridVisibilityChanged != null)
                    GridVisibilityChanged(this, new GridVisibilityChangedEventArgs("xy", xyGrid.Visible));
            }

            if (InputState.KeyPressed(Keys.F2))
            {
                xzGrid.Visible = !xzGrid.Visible;
                if( GridVisibilityChanged != null)
                    GridVisibilityChanged(this, new GridVisibilityChangedEventArgs("xz", xzGrid.Visible));
            }

            if (InputState.KeyPressed(Keys.Home))
            {
                camera.Target = Vector3.Zero;
                if (camera.Mode == CameraMode.Perspective)
                    camera.Position = new Vector3(0, 1250, 1250);
                else
                    camera.Position = new Vector3(0, 250, 0);
            }

        }

        /*private void UpdateNodes(IEnumerable<INodeUpdateable> nodes, GameTime dt, DrawInfo di)
        {
            foreach (INodeUpdateable node in nodes)
            {
                if (node.IsActive)
                {
                    node.Update(dt);
                    INodeRenderable renderable = node as INodeRenderable;
                    if (renderable != null && renderable.IsVisible)
                    {
                        renderQueue.Enqueue(renderable);
                    }
                    UpdateNodes(node.UpdateNodes, dt, di);
                }
            }
        }*/



        private void editorControl1_MouseEnter(object sender, EventArgs e)
        {
            System.Windows.Forms.Cursor.Hide();
            this.Focus();
        }

        private void editorControl1_MouseLeave(object sender, EventArgs e)
        {
            System.Windows.Forms.Cursor.Show();
            
        }

        public void editorControl1_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            GameEngine.Helpers.Cursor.Position = new Vector2(e.X, e.Y);

        }

        protected override void Dispose(bool disposing)
        {
            controls.Remove(this);
            base.Dispose(disposing);
        }

        static float scroll = 0;


        static List<EditorControl> controls = new List<EditorControl>();
        public static System.Windows.Forms.Form Form;

        static DateTime start = DateTime.Now;

        /// <summary>
        /// Updates the loop.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        public static void UpdateLoop(object sender, EventArgs e)
        {
            while (AppStillIdle)
            {
                // Update the global timer
                XNAUtil.UpdateTimer();
                if (controls.Any(x => x.Focused))
                {
                    InputState.Update();
                    InputState.ScrollMovement = scroll;
                }
                else
                {
                    InputState.Reset();
                }
                scroll = 0;
                // if the form is not focused, save som cpu cycles by delaying a little.
                if (!Form.Focused)
                    Thread.Sleep(20);

                

                //GameTime gt = new GameTime(new TimeSpan(XNAUtil.CurrentTickCount), new TimeSpan((long)XNAUtil.TimeElapsed), new TimeSpan(XNAUtil.CurrentTickCount), new TimeSpan((long)XNAUtil.TimeElapsed));
                GameTime gt = new GameTime(DateTime.Now - start, new TimeSpan(XNAUtil.ElapsedTickCount), DateTime.Now - start, new TimeSpan(XNAUtil.ElapsedTickCount));
/*                double elapsed = XNAUtil.CurrentTickCount - ticksAtLastRender;
                double fps = GraphicsDevice.DisplayMode.RefreshRate * 2;
                double frameTime = (double)TimeSpan.TicksPerSecond / (double)fps;

                if (elapsed < frameTime)
                {
                    Thread.Sleep((int)((frameTime - elapsed) / (double)TimeSpan.TicksPerMillisecond));
                }
                ticksAtLastRender = XNAUtil.CurrentTickCount;*/

                foreach (EditorControl ec in controls)
                {
                    if (ec.GraphicsDevice == null || !ec.Active)
                        continue;
                    ec.Update(gt);

                    string error = ec.BeginDraw();
                    if (error == null)
                    {
                        // This part ensures we dont render too many frames, caps the cpu usage, who wants a mediacenter that uses 100% cpu when doing nothing?

                        ec.GraphicsDevice.Clear(Microsoft.Xna.Framework.Graphics.Color.Black);

                        ec.Draw(gt);


                        Microsoft.Xna.Framework.Rectangle sourceRectangle = new Microsoft.Xna.Framework.Rectangle(0, 0, ec.ClientSize.Width, ec.ClientSize.Height);

                        try
                        {
                            ec.GraphicsDevice.Present(sourceRectangle, null, ec.Handle);
                        }
                        catch (DeviceLostException)
                        {

                        }
                    }
                    else
                    {
                    }

                }

            }
        }

        #region Application gejl

        private static bool AppStillIdle
        {
            get
            {
                PeekMsg msg;
                return !PeekMessage(out msg, IntPtr.Zero, 0, 0, 0);
            }
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct PeekMsg
        {
            public IntPtr hWnd;
            public System.Windows.Forms.Message msg;
            public IntPtr wParam;
            public IntPtr lParam;
            public uint time;
            public System.Drawing.Point p;
        }

        [SuppressUnmanagedCodeSecurity]
        [DllImport("User32.dll", CharSet = CharSet.Auto)]
        private static extern bool PeekMessage(out PeekMsg msg, IntPtr hWnd, uint messageFilterMin, uint messageFilterMax, uint flags);

        #endregion

    }
}
