using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using GameEngine.Level;
using Microsoft.Xna.Framework.Graphics;
using Editor.Tools;
using Microsoft.Xna.Framework.Content;
using GameEngine;
using GameEngine.ScreenSystem;
using Microsoft.Xna.Framework.Input;
using GameEngine.Xml;
using GameEngine.Cameras;

namespace Editor
{
    public enum OrientationPlane { XY, XZ, YZ };
    public enum Axis { X, Y, Z };

    public abstract class Tool
    {
        public class HelperGridVisibilityChangedEventArgs : EventArgs
        {
            private bool newValue;
            public bool NewValue { get { return this.newValue; } }

            public HelperGridVisibilityChangedEventArgs(bool newValue)
            {
                this.newValue = newValue;
            }
        }
        public delegate void HelperGridVisibilityChangedHandler(object sender, HelperGridVisibilityChangedEventArgs e);

        public Controls enabledControls = Controls.All;
        protected OrthoOrbitCamera camera;
        protected ToolHandler toolHandler;
        public string Name = "UnnamedTool";

        public Tool(OrthoOrbitCamera camera, ToolHandler handler)
        {
            this.camera = camera;
            this.toolHandler = handler;
        }

        abstract public void LoadContent(ContentManager content);

        virtual public void HandleInput(GameTime gameTime)
        {
            EnableControls(enabledControls);
        }
        virtual public void Update(GameTime gameTime) { }
        virtual public void Draw(DrawInfo drawInfo) { }
        virtual public void Draw2D(SpriteBatch spriteBatch, DrawInfo drawInfo) { }
        virtual public void ToolStart() { }
        virtual public void ToolEnd() { }

        [Flags]
        public enum Controls
        {
            CameraXYZ = 1,
            CameraOrbit = 2,
            MousePanning = 4,
            CameraToggle = 8,
            MouseScroll = 16,
            ToolShortcuts = 32,

            Camera = 1 | 2 | 8,
            Mouse = 4 | 16,
            All = 1024 - 1
        }

        public void CameraXYZ()
        {
            if (InputState.KeyboardState.IsKeyDown(Keys.Left))
                camera.Position.X += 1;

            if (InputState.KeyboardState.IsKeyDown(Keys.Right))
                camera.Position.X -= 1;

            if (InputState.KeyboardState.IsKeyDown(Keys.Down))
                camera.Position.Y -= 1;

            if (InputState.KeyboardState.IsKeyDown(Keys.Up))
                camera.Position.Y += 1;
        }

        public void CameraOrbit()
        {
            float speed = 0.03F;

            if (InputState.KeyboardState.IsKeyDown(Keys.A))
            {
                camera.Orbit(-speed, 0);
            }
            if (InputState.KeyboardState.IsKeyDown(Keys.D))
            {
                camera.Orbit(speed, 0);
            }
            if (InputState.KeyboardState.IsKeyDown(Keys.W))
            {
                camera.Orbit(0, -speed);
            }
            if (InputState.KeyboardState.IsKeyDown(Keys.S))
            {
                camera.Orbit(0, speed);
            }
        }

        public void MousePanning()
        {
            if (InputState.MouseState.MiddleButton == ButtonState.Pressed)
            {
                if (InputState.KeyboardState.IsKeyDown(Keys.LeftAlt))
                {
                    OrthoOrbitCamera ooc = camera as OrthoOrbitCamera;
                    ObjectNode on = toolHandler.Selection.GetFirstSelected3D();
                    if (on == null)
                    {
                        ooc.Position = OrthoOrbitCamera.Orbit(ooc.Position, ooc.Target, -InputState.MouseMovement.X * 0.01f, -InputState.MouseMovement.Y * 0.01f);
                    }
                    else
                    {
                        ooc.Position = OrthoOrbitCamera.Orbit(ooc.Position, on.WorldPosition, -InputState.MouseMovement.X * 0.01f, -InputState.MouseMovement.Y * 0.01f);
                        ooc.Target = OrthoOrbitCamera.Orbit(ooc.Target, on.WorldPosition, -InputState.MouseMovement.X * 0.01f, -InputState.MouseMovement.Y * 0.01f);
                    }

                }
                else
                {
                    Vector3 move = new Vector3(InputState.MouseMovement.X * -1f, InputState.MouseMovement.Y * 1f, 0);
                    camera.PushInDirection(move);
                }
            }
        }

        public void CameraToggle()
        {
            if (InputState.KeyPressed(Keys.Back))
            {
                if (camera.Mode == CameraMode.Perspective)
                {
                    camera.SetMode(CameraMode.Orthographic);
                    camera.Target = Vector3.Zero;
                    camera.Position = new Vector3(0, 250, 0);
                    camera.Up = -Vector3.UnitZ;
                }
                else
                {
                    camera.SetMode(CameraMode.Perspective);
                    camera.Target = new Vector3(0, 0, 125);
                    camera.Up = Vector3.Up;
                    camera.Position = new Vector3(0, 1250, 1250);
                }
            }
        }

        public void MouseScroll()
        {
            if (InputState.ScrollMovement != 0)
            {
                camera.Zoom(InputState.ScrollMovement * 0.5f);
            }
        }

        public void ToolShortcuts()
        {
            if (InputState.KeyReleased(Keys.D1))
            {
                toolHandler.SelectTool(0);
            }
            if (InputState.KeyReleased(Keys.D2))
            {
                toolHandler.SelectTool(1);
            }
            if (InputState.KeyReleased(Keys.D3))
            {
                toolHandler.SelectTool(2);
            }
            if (InputState.KeyReleased(Keys.D4))
            {
                toolHandler.SelectTool(3);
            }
            if (InputState.KeyReleased(Keys.D5))
            {
                toolHandler.SelectTool(4);
            }
            if (InputState.KeyReleased(Keys.D6))
            {
                toolHandler.SelectTool(5);
            }
            if (InputState.KeyReleased(Keys.D7))
            {
                toolHandler.SelectTool(6);
            }
            if (InputState.KeyReleased(Keys.D8))
            {
                toolHandler.SelectTool(7);
            }
            if (InputState.KeyReleased(Keys.D9))
            {
                toolHandler.SelectTool(8);
            }
            if (InputState.KeyReleased(Keys.D0))
            {
                toolHandler.SelectTool(9);
            }
        }


        public void EnableControls(Controls controls)
        {
            EnableControls(controls, false);
        }

        public void EnableControls(Controls controls, bool exclusive)
        {
            if (exclusive)
            {
                controls = ~controls;
            }

            if ((controls & Controls.CameraXYZ) == Controls.CameraXYZ)
                CameraXYZ();
            if ((controls & Controls.CameraOrbit) == Controls.CameraOrbit)
                CameraOrbit();
            if ((controls & Controls.MousePanning) == Controls.MousePanning)
                MousePanning();
            if ((controls & Controls.CameraToggle) == Controls.CameraToggle)
                CameraToggle();
            if ((controls & Controls.MouseScroll) == Controls.MouseScroll)
                MouseScroll();
            if ((controls & Controls.ToolShortcuts) == Controls.ToolShortcuts)
                ToolShortcuts();
        }



    }
}
