using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using GameEngine.Animation;
using System.Linq;
using Microsoft.Xna.Framework;
using GameEngine;
namespace GameContentPipeline
{
    [ContentProcessor(DisplayName = "Custom Animation Processor")]
    public class GameAnimationProcessor : BBoxProcessor
    {
        private const int MaxBones = 59;

        AnimationData data;

        public override ModelContent Process(NodeContent input, ContentProcessorContext context)
        {
            bool hasSkinning = HasMeshWithWeights(input);

            // Convert each animation in turn.
            data = new AnimationData() { Skinned = hasSkinning };

            ModelContent content;
            
            if (hasSkinning)
            {
                GetSkinnedAnimation(input, null);
                content = base.Process(input, context);
                if (data.Tracks.Count == 0)
                {
                    context.Logger.LogWarning("", new ContentIdentity(context.OutputFilename), "Mesh is skinned but has no animation?");
                }
                
                data.ModelBoneToSkinMap = new int[content.Bones.Count];

                for (int i = 0; i < content.Bones.Count; i++)
                {
                    if (content.Bones[i].Name == null) continue;

                    if (data.BoneMap.ContainsKey(content.Bones[i].Name))
                        data.ModelBoneToSkinMap[i] = data.BoneMap[content.Bones[i].Name];
                    else
                        data.ModelBoneToSkinMap[i] = -1;
                }
            }
            else 
            {
                content = base.Process(input, context);
                FindRigidAnimations(input, content.Bones);
            }
            
            // sort animations and fix duration.
            foreach (var animation in data.Tracks.Values)
            {
                
                animation.Keyframes.Sort((x, y) => x.Time.CompareTo(y.Time));
                animation.Duration = animation.Keyframes.Last().Time;
                
            }

            Dictionary<string, object> tag = content.Tag as Dictionary<string, object>;
            if(tag != null)
                tag.Add("AnimationData", data); 
            
            return content;
        }



        private void FindRigidAnimations(NodeContent content, ModelBoneContentCollection bones)
        {
            foreach (var anim in content.Animations)
            {
                List<Keyframe> keyframes = new List<Keyframe>();
                
                foreach (var channel in anim.Value.Channels)
                {
                    var bone = bones.FirstOrDefault(x => x.Name == channel.Key);
                    if (bone == null)
                        continue;

                    foreach (var keyframe in channel.Value)
                    {
                        // Convert the keyframe data.
                       keyframes.Add(new Keyframe(bone.Index, keyframe.Time, keyframe.Transform));
                    }
                }

                if (keyframes.Count == 0)
                    continue;

                if (!data.Tracks.ContainsKey(anim.Key))
                    data.Tracks.Add(anim.Key, new AnimationTrack(anim.Value.Duration, keyframes));
                else
                    data.Tracks[anim.Key].Keyframes.AddRange(keyframes);
            }
            
            foreach (var subNode in content.Children)
            {
                FindRigidAnimations(subNode, bones);
            }
        }

        #region SkinnedAnimation

        public AnimationData GetSkinnedAnimation(NodeContent input, ModelContent model)
        {
            BoneContent skeleton = MeshHelper.FindSkeleton(input);
            
            // Prebake transformations
            FlattenTransforms(input, skeleton);

            // Get list of bones
            List<BoneContent> skeletonBones = (List<BoneContent>)MeshHelper.FlattenSkeleton(skeleton);

            // Too many bones to handle
            if (skeletonBones.Count > MaxBones)
            {
                throw new InvalidContentException(string.Format(
                    "Skeleton has {0} bones, but the maximum supported is {1}.",
                    skeletonBones.Count, MaxBones));
            }

            Matrix[] bindPose = new Matrix[skeletonBones.Count];
            Matrix[] inverseBindPose = new Matrix[skeletonBones.Count];
            int[] skeletonHierarchy = new int[skeletonBones.Count];

            for (int i = 0; i < skeletonBones.Count; i++)
            {
                bindPose[i] = skeletonBones[i].Transform;
                inverseBindPose[i] = Matrix.Invert(skeletonBones[i].AbsoluteTransform);
                skeletonHierarchy[i] = skeletonBones.IndexOf(skeletonBones[i].Parent as BoneContent);

                if (!string.IsNullOrEmpty(skeletonBones[i].Name))
                    data.BoneMap.Add(skeletonBones[i].Name, i);
            }

            foreach (var animation in skeleton.Animations)
            {
                AnimationTrack processed = ProcessAnimation(animation.Value);
                data.Tracks.Add(animation.Key, processed);
            }
            
            data.BindPoses = bindPose;
            data.InverseBindPoses = inverseBindPose;
            data.SkeletonHierarchy = skeletonHierarchy;
            return data;
        }

        /// <summary>
        /// Converts an intermediate format content pipeline AnimationContent
        /// object to our runtime AnimationClip format.
        /// </summary>
        private AnimationTrack ProcessAnimation(AnimationContent animation)
        {
            List<Keyframe> keyframes = new List<Keyframe>();

            // For each input animation channel.
            foreach (KeyValuePair<string, AnimationChannel> channel in animation.Channels)
            {
                // Look up what bone this channel is controlling.
                int boneIndex;

                if (!data.BoneMap.TryGetValue(channel.Key, out boneIndex))
                {
                    throw new InvalidContentException(string.Format(
                        "Found animation for bone '{0}', " +
                        "which is not part of the skeleton.", channel.Key));
                }

                // Convert the keyframe data.
                foreach (AnimationKeyframe keyframe in channel.Value)
                {
                    keyframes.Add(new Keyframe(boneIndex, keyframe.Time, keyframe.Transform));
                }
            }

            if (keyframes.Count == 0)
                throw new InvalidContentException("Animation has no keyframes.");

            return new AnimationTrack(animation.Duration, keyframes);
        }



        /// <summary>
        /// Bakes unwanted transforms into the model geometry, so everything ends up in the same coordinate system.
        /// </summary>
        private void FlattenTransforms(NodeContent node, BoneContent skeleton)
        {
            if (node == skeleton)
                return;

            foreach (NodeContent child in node.Children)
            {
                // Don't process the skeleton, because that is special.
                if (child == skeleton)
                    continue;

                // Bake the local transform into the actual geometry.
                MeshHelper.TransformScene(child, child.Transform);

                // Having baked it, we can now set the local
                // coordinate system back to identity.
                child.Transform = Matrix.Identity;

                // Recurse.
                FlattenTransforms(child, skeleton);
            }
        }

        #endregion

        protected override MaterialContent ConvertMaterial(MaterialContent material, ContentProcessorContext context)
        {
            EffectMaterialContent deferredShadingMaterial = new EffectMaterialContent();
            deferredShadingMaterial.Effect = new ExternalReference<EffectContent>(EngineFX.RenderGBuffer + ".fx");
            

            // copy the textures in the original material to the new normal mapping
            // material, if they are relevant to our renderer. The
            // LookUpTextures function has added the normal map and specular map
            // textures to the Textures collection, so that will be copied as well.
            foreach (KeyValuePair<String, ExternalReference<TextureContent>> texture in material.Textures)
            {
                if ((texture.Key == "Texture"))
                    deferredShadingMaterial.Textures.Add(texture.Key, texture.Value);
            }

            return context.Convert<MaterialContent, MaterialContent>(deferredShadingMaterial, typeof(MaterialProcessor).Name);
        }

        private static bool HasMeshWithWeights(NodeContent node)
        {
            MeshContent mesh = node as MeshContent;

            if (mesh != null)
            {
                // all geometres need to have weights
                foreach (GeometryContent geometry in mesh.Geometry)
                {
                    if (!geometry.Vertices.Channels.Contains(VertexChannelNames.Weights()))
                        return false;
                }
                return true;
            }
            foreach (NodeContent child in node.Children)
            {
                if (HasMeshWithWeights(child))
                    return true;
            }
            return false;
        }


    }
}
