﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;

namespace GameEngine.Animation
{
    public class AnimationPlayer
    {
        private AnimationTrack currentTrack;
        private TimeSpan currentTimeValue;
        private int currentKeyframe;
        private Model model;
        private bool dirty = true;

        public Matrix[] BoneTransforms;
        public Matrix[] WorldTransforms;
        public Matrix[] SkinTransforms;
        public bool Ended { get; private set; }
        public bool Looping { get; private set; }
        public List<BoneManipulator> BoneManipulators { get; set; }
        public bool Pause { get; set; }
        public AnimationData Data { get; private set; }
        public string RunningAnimation { get; set; }

        public AnimationPlayer(AnimationData data, Model model)
        {
            this.Data = data;
            if (!data.Skinned)
            {
                this.BoneTransforms = new Matrix[model.Bones.Count];
                this.WorldTransforms = new Matrix[model.Bones.Count];
                if (data.ModelBoneToSkinMap == null)
                {
                    data.ModelBoneToSkinMap = new int[model.Bones.Count];
                    for (int i = 0; i < model.Bones.Count; i++)
                    {
                        data.ModelBoneToSkinMap[i] = -1;
                    }
                }
            }
            else
            {
                int c = 0, g = -1;
                if (data.SkinBoneToModelMap == null)
                {
                    data.SkinBoneToModelMap = new int[data.BindPoses.Length];
                    for (int i = 0; i < model.Bones.Count; i++)
                        if (data.ModelBoneToSkinMap[i] > -1 && g != data.ModelBoneToSkinMap[i])
                        {
                            data.SkinBoneToModelMap[c++] = i;
                            g = data.ModelBoneToSkinMap[i];
                        }
                }
                this.BoneTransforms = new Matrix[model.Bones.Count];
                this.WorldTransforms = new Matrix[model.Bones.Count];
                this.SkinTransforms = new Matrix[data.BindPoses.Length];
            }
            this.model = model;
            BoneManipulators = new List<BoneManipulator>();
            ResetPose();
        }

        private void ResetPose()
        {
            for (int i = 0; i < model.Bones.Count; i++)
            {
                if (Data.Skinned && Data.ModelBoneToSkinMap[i] > -1)
                {
                    BoneTransforms[i] = Data.BindPoses[Data.ModelBoneToSkinMap[i]];
                }
                else
                {
                    BoneTransforms[i] = model.Bones[i].Transform;
                }
            }
        }


        /// <summary>
        /// Advances the current animation position.
        /// </summary>
        public void Update(TimeSpan time, bool relativeToCurrentTime)
        {
            if (BoneManipulators.Count == 0 && Pause)
                return;
            UpdateBoneTransforms(Pause ? TimeSpan.Zero : time, relativeToCurrentTime);
            if (dirty)
            {
                UpdateWorldTransforms();
                if (Data.Skinned)
                    UpdateSkinTransforms();
                dirty = false;
            }
        }

        public void StartAnimation(string name, bool loop)
        {
            this.Looping = loop;
            this.Ended = false;
            currentTrack = Data.Tracks[name];
            currentTimeValue = TimeSpan.Zero;
            currentKeyframe = 0;
            RunningAnimation = name;
            ResetPose();
        }

        public void Reset()
        {
            RunningAnimation = null;
            currentTrack = null;
            currentTimeValue = TimeSpan.Zero;
            currentKeyframe = 0;
            ResetPose();
        }

        /// <summary>
        /// Helper used by the Update method to refresh the BoneTransforms data.
        /// </summary>
        public void UpdateBoneTransforms(TimeSpan time, bool relativeToCurrentTime)
        {
            if (currentTrack == null)
                return;

            // Update the animation position.
            if (relativeToCurrentTime)
            {
                time += currentTimeValue;

                // If we reached the end, loop back to the start.
                if (Looping)
                {
                    while (time >= (currentTrack.Duration)) time -= currentTrack.Duration;
                }
                else if (time >= currentTrack.Duration)
                {
                    Reset();
                    Ended = true;
                    return;
                }
            }

            if ((time < TimeSpan.Zero) || (time >= currentTrack.Duration))
                throw new ArgumentOutOfRangeException("time");

            // If the position moved backwards, reset the keyframe index.
            if (time < currentTimeValue)
            {
                currentKeyframe = 0;
                ResetPose();
            }

            currentTimeValue = time;

            // Read keyframe matrices.
            IList<Keyframe> keyframes = currentTrack.Keyframes;


            while (currentKeyframe < keyframes.Count)
            {
                Keyframe keyframe = keyframes[currentKeyframe];

                // Stop when we've read up to the current time position.
                if (keyframe.Time > currentTimeValue)
                    break;

                dirty = true;

                // Use this keyframe.
                if(Data.Skinned)
                    BoneTransforms[Data.SkinBoneToModelMap[keyframe.Bone]] = keyframe.Transform;
                else
                    BoneTransforms[keyframe.Bone] = keyframe.Transform;

                currentKeyframe++;
            }


        }

        /// <summary>
        /// Helper used by the Update method to refresh the WorldTransforms data.
        /// </summary>
        public void UpdateWorldTransforms()
        {
            
            for (int i = 0; i < WorldTransforms.Length; i++)
            {
                int si = Data.ModelBoneToSkinMap[i];

                if (si == -1)
                {
                    if (model.Bones[i].Parent == null)
                    {
                        WorldTransforms[i] = GetBoneTransformationWithManipulation(i);
                    }
                    else
                    {
                        WorldTransforms[i] = GetBoneTransformationWithManipulation(i) * WorldTransforms[model.Bones[i].Parent.Index];
                    }
                }
                else
                {
                    if (si == 0)
                        WorldTransforms[i] = GetBoneTransformationWithManipulation(si);
                    else
                    {
                        int parentBone = Data.SkeletonHierarchy[si];

                        WorldTransforms[i] = GetBoneTransformationWithManipulation(i) * WorldTransforms[Data.SkinBoneToModelMap[parentBone]];
                    }
                }
            }
            

            

            /*
            if (!Data.Skinned)
            {
                for (int bone = 0; bone < WorldTransforms.Length; bone++)
                {
                    if (model.Bones[bone].Parent == null)
                    {
                        WorldTransforms[bone] = BoneTransforms[bone] * rootTransform;
                    }
                    else
                    {
                        WorldTransforms[bone] = BoneTransforms[bone] * WorldTransforms[model.Bones[bone].Parent.Index];

                    }
                }
            }
            else
            {
                // Root bone.
                WorldTransforms[0] = GetBoneTransformationWithManipulation(0) * rootTransform;

                // Child bones.
                for (int bone = 1; bone < SkinTransforms.Length; bone++)
                {
                    int parentBone = Data.SkeletonHierarchy[bone];
                    WorldTransforms[bone] = GetBoneTransformationWithManipulation(bone) * WorldTransforms[parentBone];
                }
                BoneManipulators.Clear();
            }*/
        }

        private Matrix GetBoneTransformationWithManipulation(int boneIndex)
        {
            Matrix manipTransform = BoneTransforms[boneIndex];

            for (int i = 0; i < BoneManipulators.Count; i++)
            {
                if (Data.SkinBoneToModelMap[this.Data.BoneMap[BoneManipulators[i].BoneName]] == boneIndex)
                {
                    if (BoneManipulators[i].Replace)
                        manipTransform = BoneManipulators[i].Transform;
                    else
                        manipTransform = manipTransform * BoneManipulators[i].Transform;
                }
            }
            return manipTransform;
        }

        /// <summary>
        /// Helper used by the Update method to refresh the SkinTransforms data.
        /// </summary>
        public void UpdateSkinTransforms()
        {
            for (int bone = 0; bone < SkinTransforms.Length; bone++)
            {
                SkinTransforms[bone] = Data.InverseBindPoses[bone] * WorldTransforms[Data.SkinBoneToModelMap[bone]];
            }
        }
    }
}
