﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using GameEngine;
using Microsoft.Xna.Framework.Content;
using DeferredShadingTutorial;
using GameEngine.Xml;

namespace GamePrototype
{
    public class DeferredRenderer : IDisposable
    {
        private DrawInfo drawInfo;
        private Vector2 halfPixel;
        private Color bgColor = Color.TransparentBlack;
        public RenderTarget2D debugTarget;

        public RenderTarget2D colorRT, normalRT, depthRT, lightRT, ssaoRT;
        public Effect fxCombine, fxGBuffer, fxDirLight, fxPointLight, fxClearBuffer, fxSSAO;
        private Model modelSphere;
        private QuadRenderer quadRenderer;
        private Texture2D white, randomNormal;
        private SpriteBatch sb;
        private int backBufferWidth, backBufferHeight;

        public TargetOutput TargetEndResult = TargetOutput.Combined;

        public enum TargetOutput { Color, Normal, Light, Combined }

        public float Ambient { get { return fxCombine.Parameters["ambientLight"].GetValueSingle(); } set { fxCombine.Parameters["ambientLight"].SetValue(value); } }

        public DeferredRenderer(int backBufferWidth, int backBufferHeight)
        {
            this.backBufferHeight = backBufferHeight;
            this.backBufferWidth = backBufferWidth;
        }

        public void SetBGColor(Color v)
        {
            fxClearBuffer.Parameters["bgColor"].SetValue(v.ToVector4());
        }

        public void LoadContent(GraphicsDevice GraphicsDevice, ContentManager Content)
        {
            halfPixel.X = 0.5f / (float)backBufferWidth;
            halfPixel.Y = 0.5f / (float)backBufferHeight;

            // Load render targets
            colorRT = new RenderTarget2D(GraphicsDevice, backBufferWidth, backBufferHeight, 1, SurfaceFormat.Color);
            normalRT = new RenderTarget2D(GraphicsDevice, backBufferWidth, backBufferHeight, 1, SurfaceFormat.Color);
            lightRT = new RenderTarget2D(GraphicsDevice, backBufferWidth, backBufferHeight, 1, SurfaceFormat.Color);
            depthRT = new RenderTarget2D(GraphicsDevice, backBufferWidth, backBufferHeight, 1, SurfaceFormat.Single);
            ssaoRT = new RenderTarget2D(GraphicsDevice, backBufferWidth, backBufferHeight, 1, SurfaceFormat.Color);

            // Load effects
            fxClearBuffer = Content.Load<Effect>(EngineFX.ClearGBuffer);
            fxPointLight = Content.Load<Effect>(EngineFX.PointLight);
            fxDirLight = Content.Load<Effect>(EngineFX.DirectionalLight);
            modelSphere = Content.Load<Model>(EngineFX.PointLightSphere);
            fxGBuffer = Content.Load<Effect>(EngineFX.RenderGBuffer);
            //fxSSAO = Content.Load<Effect>(@"enginefx\ssao");
            fxCombine = Content.Load<Effect>(EngineFX.CombineFinal);
            fxCombine.Parameters["ambientLight"].SetValue(0f);

            // Load Helpers
            quadRenderer = new QuadRenderer();
            quadRenderer.LoadContent(GraphicsDevice);
            white = new Texture2D(GraphicsDevice, 1, 1);
            white.SetData(new Color[] { Color.White });
            randomNormal = Content.Load<Texture2D>(@"enginefx\randomnormal");
            
            sb = new SpriteBatch(GraphicsDevice);
        }


        public void Dispose()
        {
            if (white != null)
            {
                white.Dispose();
                white = null;
            }

            Util.NullDispose(ref colorRT);
            Util.NullDispose(ref normalRT);
            Util.NullDispose(ref lightRT);
            Util.NullDispose(ref depthRT);
            Util.NullDispose(ref sb);
        }


        public void BeginRender(DrawInfo di)
        {
            this.drawInfo = di;

            SetGBuffer();
            //GraphicsDevice.Clear(Color.Gray);
            ClearGBuffer();
            //scene.DrawScene(camera, gameTime);

            GraphicsDevice.RenderState.DepthBufferEnable = true;
            GraphicsDevice.RenderState.CullMode = CullMode.CullCounterClockwiseFace;
            GraphicsDevice.RenderState.AlphaBlendEnable = false;

            fxGBuffer.CurrentTechnique = fxGBuffer.Techniques[0];

            fxGBuffer.Parameters["View"].SetValue(di.Camera.View);
            fxGBuffer.Parameters["Projection"].SetValue(di.Camera.Projection);
        }

        public void EndRender(RenderTarget2D finalTarget)
        {
            /*
            for (int i = 0; i < drawInfo.PointLights.Count; i++)
            {
                GGDrawPointLight(drawInfo.PointLights[i].LightPosition, drawInfo.PointLights[i].Color, drawInfo.PointLights[i].LightRadius, drawInfo.PointLights[i].LightIntensity);
            }
            */
            
            ResolveGBuffer();
            
            GraphicsDevice.RenderState.DepthBufferEnable = false;

            if (TargetEndResult == TargetOutput.Combined || TargetEndResult == TargetOutput.Light)
            {
                GraphicsDevice.SetRenderTarget(0, lightRT);

                //clear all components to 0
                GraphicsDevice.Clear(Color.TransparentBlack);
                GraphicsDevice.RenderState.AlphaBlendEnable = true;

                //use additive blending, and make sure the blending factors are as we need them
                GraphicsDevice.RenderState.AlphaBlendOperation = BlendFunction.Add;
                GraphicsDevice.RenderState.SourceBlend = Blend.One;
                GraphicsDevice.RenderState.DestinationBlend = Blend.One;

                //use the same operation on the alpha channel
                GraphicsDevice.RenderState.SeparateAlphaBlendEnabled = false;

                //draw some lights
                //DrawDirectionalLight(new Vector3(0, -1, 1), Color.White,di );

                BoundingFrustum bf = drawInfo.Camera.Frustrum;

                for (int i = 0; i < drawInfo.PointLights.Count; i++)
                {
                    BoundingSphere bs = new BoundingSphere(drawInfo.PointLights[i].LightPosition, drawInfo.PointLights[i].LightRadius);
                    if (bf.Intersects(bs))
                        DrawPointLight(drawInfo.PointLights[i].LightPosition, drawInfo.PointLights[i].Color, drawInfo.PointLights[i].LightRadius, drawInfo.PointLights[i].LightIntensity);
                }

                for (int i = 0; i < drawInfo.UpdatePointLights.Count; i++)
                {
                    BoundingSphere bs = new BoundingSphere(drawInfo.UpdatePointLights[i].LightPosition, drawInfo.UpdatePointLights[i].LightRadius);
                    if (bf.Intersects(bs))
                        DrawPointLight(drawInfo.UpdatePointLights[i].LightPosition, drawInfo.UpdatePointLights[i].Color, drawInfo.UpdatePointLights[i].LightRadius, drawInfo.UpdatePointLights[i].LightIntensity);
                }



                GraphicsDevice.RenderState.AlphaBlendEnable = false;

                //SSAO();
                /*
                GraphicsDevice.SetRenderTarget(0, ssaoRT);
                //set the effect parameters
                fxCombine.Parameters["colorMap"].SetValue(colorRT.GetTexture());
                fxCombine.Parameters["lightMap"].SetValue(lightRT.GetTexture());
                fxCombine.Parameters["halfPixel"].SetValue(halfPixel);
                fxCombine.Begin();
                fxCombine.Techniques[0].Passes[0].Begin();
                //render a full-screen quad
                quadRenderer.Render(GraphicsDevice, Vector2.One * -1, Vector2.One);
                fxCombine.Techniques[0].Passes[0].End();
                fxCombine.End();*/


                GraphicsDevice.SetRenderTarget(0, finalTarget);

                if (TargetEndResult == TargetOutput.Combined)
                {
                    /*

                    Texture2D t = normalRT.GetTexture();
                    Texture2D t1 = colorRT.GetTexture();
                    Texture2D t2 = lightRT.GetTexture();
                    sb.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.SaveState);
                    sb.Draw(ssaoRT.GetTexture(), new Rectangle(0, 0, t.Width, t.Height), Color.White);
                    sb.Draw(t, new Rectangle(t.Width, 0, t.Width, t.Height), Color.White);
                    sb.Draw(t1, new Rectangle(0, t1.Height, t1.Width, t1.Height), Color.White);
                    sb.Draw(t2, new Rectangle(t2.Width, t2.Height, t2.Width, t2.Height), Color.White);
                    sb.End();
                    */

                    //set the effect parameters
                    fxCombine.Parameters["colorMap"].SetValue(colorRT.GetTexture());
                    fxCombine.Parameters["lightMap"].SetValue(lightRT.GetTexture());
                    fxCombine.Parameters["halfPixel"].SetValue(halfPixel);
                    fxCombine.Begin();
                    fxCombine.Techniques[0].Passes[0].Begin();
                    //render a full-screen quad
                    quadRenderer.Render(GraphicsDevice, Vector2.One * -1, Vector2.One);
                    fxCombine.Techniques[0].Passes[0].End();
                    fxCombine.End();
                }
                else
                {
                    Texture2D t = lightRT.GetTexture();
                    sb.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.SaveState);
                    sb.Draw(t, new Rectangle(0, 0, t.Width, t.Height), Color.White);
                    sb.End();
                }
            }
            else
            {
                GraphicsDevice.SetRenderTarget(0, finalTarget);
                Texture2D t = null;
                switch (TargetEndResult)
                {
                    case TargetOutput.Color:
                        t = colorRT.GetTexture();
                        break;
                    case TargetOutput.Normal:
                        t = normalRT.GetTexture();
                        break;
                }
                sb.Begin(SpriteBlendMode.None, SpriteSortMode.Immediate, SaveStateMode.SaveState);
                sb.Draw(t, new Rectangle(0, 0, t.Width, t.Height), Color.White);
                sb.End();
            }
        }

        private void SSAO()
        {
            GraphicsDevice.SetRenderTarget(0, ssaoRT);

            fxSSAO.Parameters["randomTexture"].SetValue(randomNormal);
            fxSSAO.Parameters["normalMap"].SetValue(normalRT.GetTexture());
            fxSSAO.Parameters["depthMap"].SetValue(depthRT.GetTexture());
            fxSSAO.Parameters["halfPixel"].SetValue(halfPixel);
            fxSSAO.Begin();
            fxSSAO.Techniques[0].Passes[0].Begin();
            //render a full-screen quad
            quadRenderer.Render(GraphicsDevice, Vector2.One * -1, Vector2.One);

            fxSSAO.Techniques[0].Passes[0].End();
            fxSSAO.End();

        }

        private void SetGBuffer()
        {
            GraphicsDevice.SetRenderTarget(0, colorRT);
            GraphicsDevice.SetRenderTarget(1, normalRT);
            GraphicsDevice.SetRenderTarget(2, depthRT);
        }
        public void ResolveGBuffer()
        {
            GraphicsDevice.SetRenderTarget(0, null);
            GraphicsDevice.SetRenderTarget(1, null);
            GraphicsDevice.SetRenderTarget(2, null);
        }
        private void ClearGBuffer()
        {
            fxClearBuffer.Begin();
            fxClearBuffer.Techniques[0].Passes[0].Begin();
            quadRenderer.Render(GraphicsDevice, Vector2.One * -1, Vector2.One);
            fxClearBuffer.Techniques[0].Passes[0].End();
            fxClearBuffer.End();
        }
        
        private void GGDrawPointLight(Vector3 lightPosition, Color color, float lightRadius, float lightIntensity)
        {
            //compute the light world matrix
            //scale according to light radius, and translate it to light position
            Matrix sphereWorldMatrix = Matrix.CreateScale(lightRadius * 0.01f) * Matrix.CreateTranslation(lightPosition);


            fxGBuffer.Parameters["World"].SetValue(sphereWorldMatrix);

            fxGBuffer.Begin();
            fxGBuffer.Techniques[0].Passes[0].Begin();

            foreach (ModelMesh mesh in modelSphere.Meshes)
            {
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    GraphicsDevice.VertexDeclaration = meshPart.VertexDeclaration;
                    GraphicsDevice.Vertices[0].SetSource(mesh.VertexBuffer, meshPart.StreamOffset, meshPart.VertexStride);
                    GraphicsDevice.Indices = mesh.IndexBuffer;

                    GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList,
                    meshPart.BaseVertex, 0, meshPart.NumVertices, meshPart.StartIndex,
                    meshPart.PrimitiveCount);
                }
            }

            fxGBuffer.Techniques[0].Passes[0].End();
            fxGBuffer.End();

            //GraphicsDevice.RenderState.CullMode = CullMode.CullCounterClockwiseFace;
        }




        private void DrawPointLight(Vector3 lightPosition, Color color, float lightRadius, float lightIntensity)
        {


            //set the G-Buffer parameters
            fxPointLight.Parameters["colorMap"].SetValue(colorRT.GetTexture());

            fxPointLight.Parameters["normalMap"].SetValue(normalRT.GetTexture());

            fxPointLight.Parameters["depthMap"].SetValue(depthRT.GetTexture());

            //compute the light world matrix
            //scale according to light radius, and translate it to light position
            Matrix sphereWorldMatrix = Matrix.CreateScale(lightRadius * 0.02f) * Matrix.CreateTranslation(lightPosition);

            fxPointLight.Parameters["World"].SetValue(sphereWorldMatrix);
            fxPointLight.Parameters["View"].SetValue(drawInfo.Camera.View);

            fxPointLight.Parameters["Projection"].SetValue(drawInfo.Camera.Projection);
            //light position

            fxPointLight.Parameters["lightPosition"].SetValue(lightPosition);

            //set the color, radius and Intensity
            fxPointLight.Parameters["Color"].SetValue(color.ToVector3());

            fxPointLight.Parameters["lightRadius"].SetValue(lightRadius);

            fxPointLight.Parameters["lightIntensity"].SetValue(lightIntensity);

            //parameters for specular computations

            fxPointLight.Parameters["cameraPosition"].SetValue(drawInfo.Camera.Position);

            fxPointLight.Parameters["InvertViewProjection"].SetValue(Matrix.Invert(drawInfo.Camera.View * drawInfo.Camera.Projection));
            //size of a halfpixel, for texture coordinates alignment
            fxPointLight.Parameters["halfPixel"].SetValue(halfPixel);

            //calculate the distance between the camera and light center
            float cameraToCenter = Vector3.Distance(drawInfo.Camera.Position, lightPosition);
            //if we are inside the light volume, draw the sphere's inside face
            if (cameraToCenter < lightRadius)
                GraphicsDevice.RenderState.CullMode = CullMode.CullClockwiseFace;
            else
                GraphicsDevice.RenderState.CullMode = CullMode.CullCounterClockwiseFace;

            fxPointLight.Begin();
            fxPointLight.Techniques[0].Passes[0].Begin();

            foreach (ModelMesh mesh in modelSphere.Meshes)
            {
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    GraphicsDevice.VertexDeclaration = meshPart.VertexDeclaration;
                    GraphicsDevice.Vertices[0].SetSource(mesh.VertexBuffer, meshPart.StreamOffset, meshPart.VertexStride);
                    GraphicsDevice.Indices = mesh.IndexBuffer;

                    GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList,
                    meshPart.BaseVertex, 0, meshPart.NumVertices, meshPart.StartIndex,
                    meshPart.PrimitiveCount);
                }
            }

            fxPointLight.Techniques[0].Passes[0].End();
            fxPointLight.End();

            GraphicsDevice.RenderState.CullMode = CullMode.CullCounterClockwiseFace;
        }

        GraphicsDevice GraphicsDevice { get { return drawInfo.GraphicsDevice; } }

    }
}
