﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace GameEngine.Helpers
{
    public struct CollisionRay
    {
        private Vector2 position;
        public Vector2 Position { get { return position; } set { position = value; } }
        private Vector2 direction;
        public Vector2 Direction { get { return direction; } set { direction = value; } }

        public float AngleFromXAxis
        {
            get
            {
                return (float)Math.Atan2(-Direction.Y, Direction.X);
            }
            private set{}
        }

        public void SetAngleFromXAxis(float angle) 
        {
            Matrix rotation = Matrix.CreateFromAxisAngle(Vector3.UnitY, angle);
            Vector3 rotated = Vector3.Transform(Vector3.UnitX, rotation);
            Direction = new Vector2(rotated.X, rotated.Z);
        }

        public CollisionRay(Vector2 position, Vector2 direction)
        {
            this.position = position;
            this.direction = direction;
        }

        public CollisionRay(CollisionRay ray)
        {
            position = ray.Position;
            direction = ray.Direction;
        }


        public static CollisionRay CreateFromModelInfo(Vector3 position, Quaternion orientation)
        {
            Vector3 rotated = Vector3.Transform(Vector3.UnitX, orientation);
            return new CollisionRay(new Vector2(position.X, position.Z), new Vector2(rotated.X, rotated.Z));
        }

        public CollisionRay(Vector2 position, Vector2 direction, bool normalizeDirection)
        {
            this.position = position;
            this.direction = Vector2.UnitX;
            if (normalizeDirection)
                SetAndNormalizeDirection(direction);
            else
                this.direction = direction;
        }

        public void SetAndNormalizeDirection(Vector2 direction)
        {
            this.direction = Vector2.Normalize(direction);
        }


        public void Rotate(float radians)
        {
            SetAngleFromXAxis(AngleFromXAxis + radians);
        }

        public bool Intersects(Vector2 edgeStart, Vector2 edgeEnd, out float collisionDistance)
        {
            collisionDistance = 0;

            Vector2 startToEnd = edgeEnd - edgeStart;

            float d = DotPerp(startToEnd, Direction);
            if (Math.Abs(d) < 0.00001f)
                return false;

            Vector2 delta = Position - edgeStart;

            float t = DotPerp(delta, startToEnd) / d;

            // t is how far we have traveled along the ray. If t is negative we have travelled backwards, so no intersection.
            if (t < -0.00001f)
                return false;

            // If s is between 0 and 1 we have an intersectionpoint on the line segment. Otherwise it will be
            // either to the left or the right of it.
            float s = DotPerp(delta, Direction) / d;
            if (s < 0 || s > 1)
            {
                return false;
            }
            else
            {
                collisionDistance = t;
                return true;
            }

        }

        public bool Intersects(CollisionRay ray, out float collisionDistance)
        {
            collisionDistance = 0;

            float d = DotPerp(ray.Direction, Direction);
            if (Math.Abs(d) < 0.00001f)
                return false;

            Vector2 delta = Position - ray.Position;

            float t = DotPerp(delta, ray.Direction) / d;

            // t is how far we have traveled along the ray. If t is negative we have travelled backwards, so no intersection.
            if (t < -0.00001f)
                return false;

            // s is how we have travelled along the other ray. If it's negetive, no intersection
            float s = DotPerp(delta, Direction) / d;
            if (s < -0.00001f)
            {
                return false;
            }
            else
            {
                collisionDistance = t;
                return true;
            }

        }

        public bool Intersects(BoundingCircle circle, out float collisionDistance)
        {
            collisionDistance = 0;

            Vector2 delta = Position - circle.Center;

            float directionDeltaDot = Vector2.Dot(Direction, delta);

            float d = directionDeltaDot * directionDeltaDot - Vector2.Dot(delta, delta) + circle.Radius * circle.Radius;

            // If d is negative the ray misses the circle
            if (d < -0.00001f)
                return false;

            // If d is close to 0 the ray is a tangent to the cirle
            if (d < 0.00001f)
            {
                float t = -directionDeltaDot;

                // If t is negative we collide "behind" the ray
                if (t < 0.00001f)
                {
                    return false;
                }
                else
                {
                    collisionDistance = t;
                    return true;
                }
            }
            else
            {
                float t0 = -Vector2.Dot(Direction, delta);
                float t1 = t0 + (float)Math.Sqrt(d);
                float t2 = t0 - (float)Math.Sqrt(d);
                if (t1 < 0)
                {
                    if (t2 < 0)
                    {
                        // Both collisionpoints "behind" the ray
                        return false;
                    }
                    else
                    {
                        collisionDistance = t2;
                        return true;
                    }
                }
                else
                {
                    if (t2 < 0)
                    {
                        collisionDistance = t1;
                        return true;
                    }
                    else
                    {
                        // collision point is the closest of the two collision points
                        collisionDistance = Math.Min(t1, t2);
                        return true;
                    }
                }
            }


        }

        private float DotPerp(Vector2 v1, Vector2 v2)
        {
            return v1.X * v2.Y - v2.X * v1.Y;
        }


    }
}
