using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using GameEngine.Level;

namespace GameEngine.Helpers
{
    public class Grid
    {
        private VertexBuffer buffer;
        private VertexDeclaration decl;
        private Effect primitiveEffect;
        private int width;
        private int height;
        private float distance;
        private float halfDistance;
        private Matrix world;
        private GraphicsDevice graphicsDevice;
                        
        public Vector3 Position;
        private Vector3 Scale = Vector3.One;
        public Quaternion Rotation = Quaternion.Identity;
        public bool Visible;
    
        private Color color;
        
        public Grid(int width, int height, float distance, Color color)
        {
            this.distance = distance;
            halfDistance = distance / 2;
            this.width = width;
            this.height = height;
            this.color = color;

            Visible = true;
        }

        public void ResizeLarger()
        {
            if (distance > 2)
            {

                width *= 2;
                height *= 2;
                distance /= 2;
                halfDistance = distance / 2;
                Initialize();
                
            }
        }

        public void ResizeSmaller()
        {
            if (width > 2)
            {
                width /= 2;
                height /= 2;
                halfDistance = distance;
                distance *= 2;
                Initialize();                
            }
        }

        public void Update()
        {
            world = Matrix.CreateFromQuaternion(Rotation) * Matrix.CreateScale(Scale) * Matrix.CreateTranslation(Position);//  ;
        }

        public void Draw(DrawInfo di)
        {
            if (Visible)
            {
                di.GraphicsDevice.VertexDeclaration = decl;
                di.GraphicsDevice.Vertices[0].SetSource(buffer, 0, VertexPositionColor.SizeInBytes);
                primitiveEffect.Parameters["WorldViewProj"].SetValue((world * di.Camera.View) * di.Camera.Projection);
                primitiveEffect.Parameters["Color"].SetValue(color.ToVector4());
                primitiveEffect.Begin();

                foreach (EffectPass pass in primitiveEffect.CurrentTechnique.Passes)
                {

                    pass.Begin();
                    di.GraphicsDevice.DrawPrimitives(PrimitiveType.LineList, 0, width + height + 2);
                    pass.End();
                }

                primitiveEffect.End();   
            }                        
        }

        public void Initialize()
        {
            Initialize(this.graphicsDevice, this.primitiveEffect);
        }

        public void Initialize(GraphicsDevice graphicsDevice, Effect primitivesEffect)
        {
            Unload();
            this.primitiveEffect = primitivesEffect;
            this.graphicsDevice = graphicsDevice;

            buffer = new VertexBuffer(graphicsDevice, typeof(VertexPositionColor), width * 2 + height * 2 + 4, BufferUsage.WriteOnly);

            VertexPositionColor[] vertices = new VertexPositionColor[width * 2 + height * 2 + 4];
            float x = -((float)width / 2f) * distance;
            float y = -((float)height / 2f) * distance;
            for (int i = 0; i < width; i++)
            {
                int c = i * 2;
                vertices[c] = new VertexPositionColor(new Vector3(x + i * distance, 0, y), color);
                vertices[c + 1] = new VertexPositionColor(new Vector3(x + i * distance, 0, y + height * distance), color);
            }
            for (int i = width; i < width + height; i++)
            {
                int c = i * 2;
                vertices[c] = new VertexPositionColor(new Vector3(x, 0, y + (i - width) * distance), color);
                vertices[c + 1] = new VertexPositionColor(new Vector3(x + width * distance, 0, y + (i - width) * distance), color);
            }

            vertices[vertices.Length - 4] = new VertexPositionColor(new Vector3(x, 0, y + height * distance), color);
            vertices[vertices.Length - 3] = new VertexPositionColor(new Vector3(x + width * distance, 0, y + height * distance), color);
            vertices[vertices.Length - 2] = new VertexPositionColor(new Vector3(x + width * distance, 0, y), color);
            vertices[vertices.Length - 1] = new VertexPositionColor(new Vector3(x + width * distance, 0, y + height * distance), color);

            buffer.SetData(vertices);

            decl = new VertexDeclaration(graphicsDevice, VertexPositionColor.VertexElements);
        }

        public void Unload()
        {
            if (buffer != null)
            {
                buffer.Dispose();
                buffer = null;
            }
            if (decl != null)
            {
                decl.Dispose();
                decl = null;
            }
        }

        public Vector2 Snap(Vector2 input)
        {
            Vector2 result = new Vector2(Snap(input.X), Snap(input.Y));
            return result;
        }

        public float Snap(float unsnapped)
        {
            float rest = unsnapped % distance;

            float result = unsnapped;
            result -= rest;

            if (Math.Abs(rest) >= halfDistance)
                result += Math.Sign(rest) * distance;

            return result;
        }
    }
}
