﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;

namespace GameEngine.Helpers
{
    public class Quad : IDisposable
    {
        VertexBuffer buffer;
        IndexBuffer indices;
        VertexDeclaration decl;
        public Vector3 Normal { get; private set; }
        public Vector3 Center { get; private set; }
        int size;
        private Quad()
        {
        }

        ~Quad()
        {
            Dispose();
        }

        public static Quad GenerateWall(GraphicsDevice device, Vector3 start, Vector3 end, Vector3 height)
        {
            return GenerateWall(device, start, end, height, Vector2.Zero, Vector2.One);

        }

        public static Quad GenerateWall(GraphicsDevice device, Vector3 start, Vector3 end, Vector3 height, Vector2 texCoordTopLeft, Vector2 texCoordBottomRight)
        {
            Quad q = new Quad();
            q.buffer = new VertexBuffer(device, typeof(VertexPositionColorTexture), 4, BufferUsage.WriteOnly);
            
            VertexPositionColorTexture[] vertices = new VertexPositionColorTexture[4] 
            {
                new VertexPositionColorTexture(start + height, Color.Red, texCoordTopLeft),
                new VertexPositionColorTexture(end + height, Color.Red, new Vector2(texCoordBottomRight.X, texCoordTopLeft.Y)),
                new VertexPositionColorTexture(end, Color.Red, texCoordBottomRight),
                new VertexPositionColorTexture(start, Color.Red,  new Vector2(texCoordTopLeft.X, texCoordBottomRight.Y))
            };
            q.buffer.SetData(vertices);

            q.indices = new IndexBuffer(device, typeof(short), 4, BufferUsage.WriteOnly);
            q.indices.SetData(new short[] { 0, 1, 3, 2 });
            q.decl = new VertexDeclaration(device, VertexPositionColorTexture.VertexElements);
            q.Normal = Vector3.Cross(height, end - start);
            q.Normal = Vector3.Normalize(q.Normal);
            q.Center = (end + start) / 2f + height / 2f;
            q.size = VertexPositionColorTexture.SizeInBytes;

            return q;

        }

        public static Quad Generate(GraphicsDevice device, Vector3 origin)
        {
            Quad q = new Quad();
            q.buffer = new VertexBuffer(device, typeof(VertexPositionNormalTexture), 4, BufferUsage.WriteOnly);

            VertexPositionNormalTexture[] vertices = new VertexPositionNormalTexture[4] 
            {
                new VertexPositionNormalTexture(new Vector3(-0.5f, -0.5f, 0), Vector3.UnitZ, Vector2.Zero),
                new VertexPositionNormalTexture(new Vector3(0.5f, -0.5f, 0), Vector3.UnitZ, Vector2.UnitX),
                new VertexPositionNormalTexture(new Vector3(0.5f, 0.5f, 0), Vector3.UnitZ, Vector2.One),
                new VertexPositionNormalTexture(new Vector3(-0.5f, 0.5f, 0), Vector3.UnitZ,  Vector2.UnitY)
            };
            q.buffer.SetData(vertices);

            q.indices = new IndexBuffer(device, typeof(short), 4, BufferUsage.WriteOnly);
            q.indices.SetData(new short[] { 0, 1, 3, 2 });
            q.decl = new VertexDeclaration(device, VertexPositionNormalTexture.VertexElements);
            q.size = VertexPositionNormalTexture.SizeInBytes;
            return q;
        }


        public static Quad GenerateXZ(GraphicsDevice device, float scale, float height, Vector2 textureOffset, Vector2 textureSize, Vector4 color) 
        {
            Quad q = new Quad();
            q.buffer = new VertexBuffer(device, typeof(VertexPositionColorTexture), 4, BufferUsage.WriteOnly);

            VertexPositionColorTexture[] vertices = new VertexPositionColorTexture[4] 
            {
                new VertexPositionColorTexture(new Vector3(-.5f * scale,height,-.5f * scale) , new Color(color), textureSize == Vector2.Zero ? Vector2.Zero : new Vector2(textureOffset.X / textureSize.X, textureOffset.Y / textureSize.Y)),
                new VertexPositionColorTexture(new Vector3(.5f * scale,height,-.5f * scale), new Color(color), textureSize == Vector2.Zero ? Vector2.UnitX : new Vector2(textureOffset.X / textureSize.X + scale / textureSize.X, textureOffset.Y / textureSize.Y)),
                new VertexPositionColorTexture(new Vector3(.5f * scale,height,.5f * scale), new Color(color), textureSize == Vector2.Zero ? Vector2.One : new Vector2(textureOffset.X / textureSize.X + scale / textureSize.X, textureOffset.Y / textureSize.Y + scale / textureSize.Y)),
                new VertexPositionColorTexture(new Vector3(-.5f * scale,height,.5f * scale), new Color(color),  textureSize == Vector2.Zero ? Vector2.UnitY : new Vector2(textureOffset.X / textureSize.X, textureOffset.Y / textureSize.Y + scale / textureSize.Y))
            };
            q.buffer.SetData(vertices);

            q.indices = new IndexBuffer(device, typeof(short), 4, BufferUsage.WriteOnly);
            q.indices.SetData(new short[] { 0, 1, 3, 2 });
            q.decl = new VertexDeclaration(device, VertexPositionColorTexture.VertexElements);
            q.size = VertexPositionColorTexture.SizeInBytes;
            return q;
        }

        public Matrix GetXZWorld(Vector3 topLeft, float width, float height)
        {
            //  + new Vector3(width / 2f,0 , height / 2f)
            return Matrix.CreateScale(new Vector3(width, 0, height)) * Matrix.CreateTranslation(topLeft);
        }

        public void SetSource(GraphicsDevice device)
        {
            device.Vertices[0].SetSource(buffer, 0, size);
            device.Indices = indices;
        }

        public void Draw(GraphicsDevice device, Effect effect)
        {
            SetSource(device);
            effect.Begin();

            for (int i = 0; i < effect.CurrentTechnique.Passes.Count; i++)
            {
                effect.CurrentTechnique.Passes[i].Begin();

                Draw(device);

                effect.CurrentTechnique.Passes[i].End();
            }

            effect.End();
        }

        public void Draw(GraphicsDevice device)
        {
            device.DrawIndexedPrimitives(PrimitiveType.TriangleStrip, 0, 0, 4, 0, 2);
        }

        public void Dispose()
        {
            Util.NullDispose(ref buffer);
            Util.NullDispose(ref indices);
            Util.NullDispose(ref decl);
        }

        
    }
}
