using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.ScreenSystem;
using Microsoft.Xna.Framework.Input;
using GameEngine.Helpers;
using GameEngine.Xml;
using FarseerGames.FarseerPhysics.Collisions;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics;

namespace GameEngine.Level
{
    public class Level
    {
        private OrthoOrbitCamera camera;
        private ContentManager content;
        private GraphicsDeviceManager graphics;
        private DrawInfo drawInfo = new DrawInfo(Matrix.Identity, Matrix.Identity);
        private SpriteFont font;
        private SpriteBatch spriteBatch;
        private Texture2D cursorTexture;

        private string levelAsset;
        private ForestHandler sceneTree;

        private Queue<INodeRenderable> renderQueue = new Queue<INodeRenderable>();

        public Level(Game game, GraphicsDeviceManager graphics, string asset)
        {
            this.levelAsset = asset;
            this.graphics = graphics;
            this.content = new ContentManager(game.Services, "Content");
        }

        
        public void LoadContent()
        {
            spriteBatch = new SpriteBatch(graphics.GraphicsDevice);

            font = content.Load<SpriteFont>("Fonts/Arial");
            cursorTexture = content.Load<Texture2D>("Hud/cursor");
            

            camera = new OrthoOrbitCamera(MathHelper.PiOver2, new Vector3(0, 0, 100), new Vector3(10, 0, 10), Vector3.Up, graphics.GraphicsDevice.Viewport, 1f, 1000f, CameraMode.Perspective);
            drawInfo.GraphicsDevice = graphics.GraphicsDevice;
            graphics.GraphicsDevice.RenderState.DepthBufferEnable = true;

            sceneTree = new ForestHandler();
            sceneTree.AddRoots(XmlTest.LoadTest(levelAsset, content));
            Dictionary<string,object> loadProperties = new Dictionary<string,object>();

            foreach (var node in sceneTree.NodesBreadthFirst)
            {
                INodeLoadable nodeLoad = node as INodeLoadable;
                if (nodeLoad != null)
                    nodeLoad.LoadContent(content, graphics.GraphicsDevice, sceneTree.Map, loadProperties);
            }            


        }

        public void Update(GameTime dt)
        {
            HandleInput(dt);

            // Update the Camera
            camera.UpdateCamera(dt);
            
            // Update the draw information
            drawInfo.Projection = camera.Projection;
            drawInfo.View = camera.View;

            // Update the forest
            foreach (var node in sceneTree.NodesBreadthFirst)
            {
                INodeUpdateable updateNode = node as INodeUpdateable;
                if (updateNode != null && updateNode.IsActive)
                {
                    updateNode.Update(dt, drawInfo);
                    
                    // if this node can be rendered, add it to the render queue if it's visible
                    INodeRenderable renderNode = node as INodeRenderable;
                    if (renderNode != null && renderNode.IsVisible)
                    {
                        renderQueue.Enqueue(renderNode);
                    }
                }
            }    
        }

        private void HandleInput(GameTime dt)
        {
            


        }

        public void Draw(GameTime dt)
        {
            // Empty the renderQueue, rendering on the way
            while (renderQueue.Count > 0)
                renderQueue.Dequeue().Draw(dt, drawInfo);

            // Draw some debug info
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            spriteBatch.DrawString(font, "Cam position:    " + camera.Position.ToString(), new Vector2(30, 10), Color.White);
            spriteBatch.End();
        }
    }
}
