using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Xml;
using GameEngine.LevelGeneration.Graphs;
using Microsoft.Xna.Framework;

namespace GameEngine.LevelGeneration
{
    public class LevelGenerator
    {
        public LevelTile[,] LevelTiles { get; set; }
        public List<LevelTile> CreatedTiles { get; set; }
        public List<LevelTile> PointsOfInterest { get; set; }

        private Dictionary<LevelTileNode, TileType> availableSceneNodes;

        private int AverageTiles;
        private int AllowedDeviation;
        private int MaxTiles;

        public LevelGenerator(Dictionary<LevelTileNode, TileType> availableSceneNodes)
        {
            this.availableSceneNodes = availableSceneNodes;

            LevelTiles = new LevelTile[2 * MaxTiles, 2 * MaxTiles];
            CreatedTiles = new List<LevelTile>();
            PointsOfInterest = new List<LevelTile>();

            this.AverageTiles = 75;
            this.AllowedDeviation = 25;
            this.MaxTiles = 100;
            ConstructLevel(75, 25);
        }

        public LevelGenerator(int averageTiles, int allowedDeviation)
        {
            this.availableSceneNodes = new Dictionary<LevelTileNode, TileType>();
            this.AverageTiles = averageTiles;
            this.AllowedDeviation = allowedDeviation;
            this.MaxTiles = (averageTiles + allowedDeviation);

            LevelTiles = new LevelTile[2 * MaxTiles, 2 * MaxTiles];
            CreatedTiles = new List<LevelTile>();
            PointsOfInterest = new List<LevelTile>();
            ConstructLevel(averageTiles, allowedDeviation);
        }

        #region Level Information

        private void GetLevelMinMax(out int minX, out int maxX, out int minY, out int maxY)
        {
            LevelTile temp;

            minX = MaxTiles * 2;
            maxX = 0;
            minY = MaxTiles * 2;
            maxY = 0;

            for (int i = 0; i < MaxTiles * 2; i++)
            {
                for (int j = 0; j < MaxTiles * 2; j++)
                {
                    temp = LevelTiles[i, j];

                    if (temp != null)
                    {
                        minX = minX > i ? i : minX;
                        maxX = maxX < i ? i : maxX;

                        minY = minY > j ? j : minY;
                        maxY = maxY < j ? j : maxY;
                    }
                }
            }
        }

        public LevelTile[,] GetLevelTileArray(out int width, out int height)
        {
            return GetLevelTileArray(out width, out height, 0);
        }

        public LevelTile[,] GetLevelTileArray(out int width, out int height, int padding)
        {
            int minX, maxX, minY, maxY;
            GetLevelMinMax(out minX, out maxX, out minY, out maxY);

            int levelWidth = (maxX - minX + 1);
            int levelHeight = (maxY - minY + 1);
            width = levelWidth + (padding * 2);
            height = levelHeight + (padding * 2);
            LevelTile[,] paddedLevel = new LevelTile[width, height];

            LevelTile temp;
            for (int i = 0; i < levelWidth; i++)
            {
                for (int j = 0; j < levelHeight; j++)
                {
                    temp = LevelTiles[i + minX, j + minY];
                    if (temp != null)
                        paddedLevel[i + padding, j + padding] = temp;
                }
            }

            return paddedLevel;
        }

        public List<LevelTile> GetLevelTileList()
        {
            List<LevelTile> allTiles = new List<LevelTile>();
            LevelTile temp;

            int minX, maxX, minY, maxY;
            GetLevelMinMax(out minX, out maxX, out minY, out maxY);

            bool horizontal = ((maxX - minX) >= (maxY - minY)) ? true : false;

            if (horizontal)
            {
                for (int i = 0; i < MaxTiles * 2; i++)
                {
                    for (int j = 0; j < MaxTiles * 2; j++)
                    {
                        temp = LevelTiles[i, j];
                        if (temp != null) { allTiles.Add(temp); }
                    }
                }
            }
            else
            {
                for (int i = 0; i < MaxTiles * 2; i++)
                {
                    for (int j = 0; j < MaxTiles * 2; j++)
                    {
                        temp = LevelTiles[j, i];
                        if (temp != null) { allTiles.Add(temp); }
                    }
                }
            }
            return allTiles;
        }
        #endregion

        #region Level construction

        public void ConstructLevel(int averageTiles, int allowedDeviation)
        {
            this.AverageTiles = averageTiles;
            this.AllowedDeviation = allowedDeviation;
            this.MaxTiles = averageTiles + allowedDeviation;
            PointsOfInterest.Clear();

            int infinityProtection = 0;

            do
            {
                // Clear out any potential existing level information
                LevelTiles = new LevelTile[2 * MaxTiles, 2 * MaxTiles];
                CreatedTiles.Clear();

                // Recursively create the new level...
                CreateTile(MaxTiles, MaxTiles);
            }
            while (CreatedTiles.Count < (averageTiles - allowedDeviation) && infinityProtection++ < 100);
        }

        public void ConstructLevel(int averageTiles, int allowedDeviation, List<LevelTile> baseLevel)
        {
            this.AverageTiles = averageTiles;
            this.AllowedDeviation = allowedDeviation;
            this.MaxTiles = averageTiles + allowedDeviation;
            PointsOfInterest.Clear();

            int infinityProtection = 0;

            do
            {
                // Clear out any potential existing level information
                LevelTiles = new LevelTile[2 * MaxTiles, 2 * MaxTiles];
                CreatedTiles.Clear();

                // Add in the tiles in the base level
                foreach (LevelTile currentTile in baseLevel)
                {
                    LevelTiles[currentTile.X + this.MaxTiles, currentTile.Y + this.MaxTiles] = currentTile;
                    CreatedTiles.Add(currentTile);
                    if (currentTile.IsPointOfInterest)
                        PointsOfInterest.Add(currentTile);
                }

                // Now create tiles according to the newly added tiles
                // Cannot do this in the same loop, as we may well end up missing certain tiles in the base level...
                foreach (LevelTile currentTile in baseLevel)
                {
                    if (currentTile.TopEdge == EdgeState.Edge && currentTile.TileAbove == null)
                        CreateTile(currentTile.X, currentTile.Y - 1);
                    if (currentTile.RightEdge == EdgeState.Edge && currentTile.TileRight == null)
                        CreateTile(currentTile.X + 1, currentTile.Y);
                    if (currentTile.BottomEdge == EdgeState.Edge && currentTile.TileBelow == null)
                        CreateTile(currentTile.X, currentTile.Y + 1);
                    if (currentTile.LeftEdge == EdgeState.Edge && currentTile.TileLeft == null)
                        CreateTile(currentTile.X - 1, currentTile.Y);
                }
            }
            while (CreatedTiles.Count < (averageTiles - allowedDeviation) && infinityProtection++ < 100);
        }

        private void CreateTile(int xPos, int yPos)
        {
            LevelTile newTile;

            if (LevelTiles[xPos, yPos] == null)
            {
                newTile = new LevelTile(xPos - MaxTiles, yPos - MaxTiles);
                LevelTiles[xPos, yPos] = newTile;
                CreatedTiles.Add(newTile);
            }
            else
                return;

            // Find forced edges in adjoining tiles
            FindForcedEdges(xPos, yPos, newTile);

            // Find an appropriate tile in our list of available tiles
            newTile.RandomizeTile(this.availableSceneNodes);

            // Which edges still don't have a defined edge (as either edge or blocked)
            List<Rotation> availableEdges = new List<Rotation>();
            if (newTile.TopEdge == EdgeState.None)
                availableEdges.Add(Rotation.Up);
            if (newTile.RightEdge == EdgeState.None)
                availableEdges.Add(Rotation.Right);
            if (newTile.BottomEdge == EdgeState.None)
                availableEdges.Add(Rotation.Down);
            if (newTile.LeftEdge == EdgeState.None)
                availableEdges.Add(Rotation.Left);

            // We might as well bail out here in stead of making a whole lot of calculations and instantiations below...
            if (availableEdges.Count == 0)
                return;

            // Shuffling
            int n = availableEdges.Count;
            Rotation temp;
            while (n > 1)
            {
                int k = Util.Random.Next(n);
                n--;
                temp = availableEdges[n];
                availableEdges[n] = availableEdges[k];
                availableEdges[k] = temp;
            }

            // How many edges do we need to worry about
            int newEdges = availableEdges.Count;

            // ...of course make sure we don't make too many tiles
            if (AverageTiles - CreatedTiles.Count < newEdges)
                newEdges = AverageTiles - CreatedTiles.Count;

            // Make sure newEdges stays above -1
            newEdges = (newEdges < 0) ? 0 : newEdges;

            // If we've been capped, we need to make sure the level actually ends somewhere!
            if (newEdges != availableEdges.Count)
            {
                newTile.TopEdge = (newTile.TopEdge == EdgeState.None) ? EdgeState.Blocked : newTile.TopEdge;
                newTile.RightEdge = (newTile.RightEdge == EdgeState.None) ? EdgeState.Blocked : newTile.RightEdge;
                newTile.BottomEdge = (newTile.BottomEdge == EdgeState.None) ? EdgeState.Blocked : newTile.BottomEdge;
                newTile.LeftEdge = (newTile.LeftEdge == EdgeState.None) ? EdgeState.Blocked : newTile.LeftEdge;
                return;
            }

            int bushiness = 50;
            // Figure out what edges should be along the tile at the available sides
            for (int i = 0; i < newEdges; i++)
            {
                switch (availableEdges[i])
                {
                    case Rotation.Up:
                        if (Util.Random.Next(100) < bushiness && CreatedTiles.Count < MaxTiles)
                        {
                            newTile.TopEdge = EdgeState.Edge;
                            newTile.TileAbove = LevelTiles[xPos, yPos - 1];
                        }
                        else
                            newTile.TopEdge = EdgeState.Blocked;
                        break;
                    case Rotation.Right:
                        if (Util.Random.Next(100) < bushiness && CreatedTiles.Count < MaxTiles)
                        {
                            newTile.RightEdge = EdgeState.Edge;
                            newTile.TileRight = LevelTiles[xPos + 1, yPos];
                        }
                        else
                            newTile.RightEdge = EdgeState.Blocked;
                        break;
                    case Rotation.Down:
                        if (Util.Random.Next(100) < bushiness && CreatedTiles.Count < MaxTiles)
                        {
                            newTile.BottomEdge = EdgeState.Edge;
                            newTile.TileBelow = LevelTiles[xPos, yPos + 1];
                        }
                        else
                            newTile.BottomEdge = EdgeState.Blocked;
                        break;
                    case Rotation.Left:
                    default:
                        if (Util.Random.Next(100) < bushiness && CreatedTiles.Count < MaxTiles)
                        {
                            newTile.LeftEdge = EdgeState.Edge;
                            newTile.TileLeft = LevelTiles[xPos - 1, yPos];
                        }
                        else
                            newTile.LeftEdge = EdgeState.Blocked;
                        break;
                }
            }
            
            // Create the new adjoining tiles
            for (int i = 0; i < newEdges; i++)
            {
                switch (availableEdges[i])
                {
                    case Rotation.Up:
                        if (newTile.TopEdge == EdgeState.Edge)
                            CreateTile(xPos, yPos - 1);
                        break;
                    case Rotation.Right:
                        if (newTile.RightEdge == EdgeState.Edge)
                            CreateTile(xPos + 1, yPos);
                        break;
                    case Rotation.Down:
                        if (newTile.BottomEdge == EdgeState.Edge)
                            CreateTile(xPos, yPos + 1);
                        break;
                    case Rotation.Left:
                    default:
                        if (newTile.LeftEdge == EdgeState.Edge)
                            CreateTile(xPos - 1, yPos);
                        break;
                }
            }

            // Ensure the edges are fully punched through
            FindForcedEdges(xPos, yPos, newTile);
        }

        private void FindForcedEdges(int x, int y, LevelTile theTile)
        {
            // if (x+/-1,y+/-1) isn't null, and it has an edge to (x,y), mark that edge in (x,y) as connected
            // if (x+/-1,y+/-1) isn't null, and does not have an edge to (x,y), mark that edge in (x,y) as blocked

            LevelTile above = LevelTiles[x, y - 1];
            if (above != null)
            {
                theTile.TileAbove = above;
                if (above.BottomEdge == EdgeState.Edge)
                    theTile.TopEdge = EdgeState.Edge;
                else if (above.BottomEdge == EdgeState.Blocked)
                    theTile.TopEdge = EdgeState.Blocked;
            }

            LevelTile right = LevelTiles[x + 1, y];
            if (right != null)
            {
                theTile.TileRight = right;
                if (right.LeftEdge == EdgeState.Edge)
                    theTile.RightEdge = EdgeState.Edge;
                else if (right.LeftEdge == EdgeState.Blocked)
                    theTile.RightEdge = EdgeState.Blocked;
            }

            LevelTile below = LevelTiles[x, y + 1];
            if (below != null)
            {
                theTile.TileBelow = below;
                if (below.TopEdge == EdgeState.Edge)
                    theTile.BottomEdge = EdgeState.Edge;
                else if (below.TopEdge == EdgeState.Blocked)
                    theTile.BottomEdge = EdgeState.Blocked;
            }

            LevelTile left = LevelTiles[x - 1, y];
            if (left != null)
            {
                theTile.TileLeft = left;
                if (left.RightEdge == EdgeState.Edge)
                    theTile.LeftEdge = EdgeState.Edge;
                else if (left.RightEdge == EdgeState.Blocked)
                    theTile.LeftEdge = EdgeState.Blocked;
            }
        }

        public void RandomizeSceneNodes()
        {
            RandomizeSceneNodes(this.availableSceneNodes);
        }

        public void RandomizeSceneNodes(Dictionary<LevelTileNode, TileType> availableNodes)
        {
            this.availableSceneNodes = availableNodes;
            foreach (LevelTile theTile in CreatedTiles)
                theTile.RandomizeTile(availableSceneNodes);
        }

        #endregion

        #region Points Of Interest

        public bool PositionPoints(int numberOfPoints)
        {
            return FindScatteredPositions(numberOfPoints);
            // This doesn't get implemented until we have pathfinding - next semester!
            //PositionPoints(numberOfPoints, CreatedTiles.Count / (numberOfPoints * 2));
        }

        private bool FindScatteredPositions(int n)
        {
            List<LevelTile> allTiles = GetLevelTileList();
            PointsOfInterest.Clear();
            for (int i = 0; i < allTiles.Count; i++)
                allTiles[i].IsPointOfInterest = false;

            int partionSize = allTiles.Count / n; 
            
            int k;
            LevelTile aTile;
            List<LevelTile> tempNodes = new List<LevelTile>();
            for (int i = 0; i < n; i++)
            {
                // First and last point of interest MUST be an end node - avoid having to make too many tiles for this...
                if (i == 0 || i == n - 1 || i == n - 2)
                {
                    tempNodes.Clear();
                    for (int j = i * partionSize; j < partionSize + i * partionSize; j++)
                        if(allTiles[j].TileType == TileType.End)
                            tempNodes.Add(allTiles[j]);

                    if (tempNodes.Count == 0)
                        return false;
                    else
                        aTile = tempNodes[Util.Random.Next(tempNodes.Count)];
                }
                else
                {
                    k = Util.Random.Next(partionSize) + i * partionSize;
                    while (k >= allTiles.Count)
                        k--;
                    aTile = allTiles[k];
                }

                aTile.IsPointOfInterest = true;
                PointsOfInterest.Add(aTile);
            }
            return true;
        }

        private bool PositionPoints(int numberOfPoints, int requiredDistance)
        {
            LevelTile theTile;
            List<LevelTile> interestingTiles = new List<LevelTile>();
            int pointCount = 0, numberOfTries = 0;
            DirectedGraph theGraph = GetTileGraph();

            bool tooClose;
            do
            {
                do
                {
                    theTile = CreatedTiles[Util.Random.Next(CreatedTiles.Count)];
                    if (!theTile.IsPointOfInterest)
                    {
                        theTile.IsPointOfInterest = true;
                        interestingTiles.Add(theTile);
                    }
                }
                while (++pointCount < numberOfPoints);

                // This is a super-nasty hack, but it makes it at least work
                // If we don't get sanely positioned points in a hundred tries,
                // we assume that the level is simply too bushy and that we thus
                // cannot place the points sanely enough - just bail out
                if (numberOfTries++ > 100)
                    break;

                // Do a distance calculation between all combinations of assigned points of interest
                // to ensure that the points are far enough apart...
                tooClose = false;

                foreach (LevelTile from in interestingTiles)
                    foreach (LevelTile to in interestingTiles)
                        if (theGraph.CalculateDistance(from, to) < requiredDistance)
                            tooClose = true;
            }
            while(tooClose);

            PointsOfInterest = interestingTiles;

            return true;
        }

        private DirectedGraph GetTileGraph()
        {
            DirectedGraph theGraph = new DirectedGraph();
            Dictionary<LevelTile, Node> nodeList = new Dictionary<LevelTile, Node>();

            // Add the nodes first
            foreach (LevelTile currentTile in CreatedTiles)
            {
                nodeList.Add(currentTile, new Node());
                theGraph.Nodes.Add(currentTile, nodeList[currentTile]);
            }

            // Now add the edges between them - this algorithm as a bonus makes sure we have both directions in all edges
            foreach (LevelTile theTile in CreatedTiles)
            {
                if (theTile.TopEdge == EdgeState.Edge)
                    theGraph.Edges.Add(new Edge(nodeList[theTile], nodeList[theTile.TileAbove], 1));
                if (theTile.RightEdge == EdgeState.Edge)
                    theGraph.Edges.Add(new Edge(nodeList[theTile], nodeList[theTile.TileRight], 1));
                if (theTile.BottomEdge == EdgeState.Edge)
                    theGraph.Edges.Add(new Edge(nodeList[theTile], nodeList[theTile.TileBelow], 1));
                if (theTile.LeftEdge == EdgeState.Edge)
                    theGraph.Edges.Add(new Edge(nodeList[theTile], nodeList[theTile.TileLeft], 1));
            }

            return theGraph;
        }
        #endregion


    }
}
