using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework;

namespace GameEngine.ScreenSystem
{
    public static class InputState
    {
        private static KeyboardState currentKeyboardState;
        private static KeyboardState lastKeyboardState;
        private static MouseState currentMouseState;
        private static MouseState lastMouseState;
        public static Vector2 MouseMovement;
        public static float ScrollMovement;

        private static GamePadState[] currentGamePadState = new GamePadState[4];
        private static GamePadState[] lastGamePadState = new GamePadState[4];

        static InputState()
        {
            currentKeyboardState = Keyboard.GetState();
            currentMouseState = Mouse.GetState();
            currentGamePadState[(int)PlayerIndex.One] = GamePad.GetState(PlayerIndex.One);
            currentGamePadState[(int)PlayerIndex.Two] = GamePad.GetState(PlayerIndex.Two);
            currentGamePadState[(int)PlayerIndex.Three] = GamePad.GetState(PlayerIndex.Three);
            currentGamePadState[(int)PlayerIndex.Four] = GamePad.GetState(PlayerIndex.Four);
        }

        public static void Reset()
        {
            currentKeyboardState = lastKeyboardState;
            currentMouseState = lastMouseState;
            MouseMovement = Vector2.Zero;
            ScrollMovement = 0;
        }

        public static void Update()
        {
            lastKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            lastMouseState = currentMouseState;
            currentMouseState = Mouse.GetState();
            ScrollMovement = currentMouseState.ScrollWheelValue - lastMouseState.ScrollWheelValue;
            MouseMovement = new Vector2(currentMouseState.X - lastMouseState.X, currentMouseState.Y - lastMouseState.Y);

            for (int i = 0; i < 4; i++)
                lastGamePadState[i] = currentGamePadState[i];
            currentGamePadState[(int)PlayerIndex.One] = GamePad.GetState(PlayerIndex.One);
            currentGamePadState[(int)PlayerIndex.Two] = GamePad.GetState(PlayerIndex.Two);
            currentGamePadState[(int)PlayerIndex.Three] = GamePad.GetState(PlayerIndex.Three);
            currentGamePadState[(int)PlayerIndex.Four] = GamePad.GetState(PlayerIndex.Four);
        }

        public static bool IsKeyDown(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key);
        }

        public static bool IsKeyUp(Keys key)
        {
            return currentKeyboardState.IsKeyUp(key);
        }

        public static Keys[] GetPressedKeys()
        {
            return currentKeyboardState.GetPressedKeys();
        }

        /// <summary>
        /// Formerly known as IsJustPressed
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public static bool KeyPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && !lastKeyboardState.IsKeyDown(key);
        }

        public static bool KeyReleased(Keys key)
        {
            return !currentKeyboardState.IsKeyDown(key) && lastKeyboardState.IsKeyDown(key);
        }

        public static bool MousePressed(MouseButton button)
        {
            switch (button)
            {
                case MouseButton.Left:
                    return currentMouseState.LeftButton == ButtonState.Pressed && lastMouseState.LeftButton == ButtonState.Released;
                case MouseButton.Right:
                    return currentMouseState.RightButton == ButtonState.Pressed && lastMouseState.RightButton == ButtonState.Released;
                case MouseButton.Middle:
                    return currentMouseState.MiddleButton == ButtonState.Pressed && lastMouseState.MiddleButton == ButtonState.Released;
            }
            return false;
        }
        public static bool MouseReleased(MouseButton button)
        {
            switch (button)
            {
                case MouseButton.Left:
                    return currentMouseState.LeftButton == ButtonState.Released && lastMouseState.LeftButton == ButtonState.Pressed;
                case MouseButton.Right:
                    return currentMouseState.RightButton == ButtonState.Released && lastMouseState.RightButton == ButtonState.Pressed;
                case MouseButton.Middle:
                    return currentMouseState.MiddleButton == ButtonState.Released && lastMouseState.MiddleButton == ButtonState.Pressed;
            }
            return false;
        }

        public static KeyboardState KeyboardState { get { return currentKeyboardState; } }
        public static MouseState MouseState { get { return currentMouseState; } }

        public static bool ButtonPressed(PlayerIndex playerIndex, Buttons button)
        {
            return currentGamePadState[(int)playerIndex].IsButtonDown(button) && !lastGamePadState[(int)playerIndex].IsButtonDown(button);
        }

        public static bool ButtonReleased(PlayerIndex playerIndex, Buttons button)
        {
            return !currentGamePadState[(int)playerIndex].IsButtonDown(button) && lastGamePadState[(int)playerIndex].IsButtonDown(button);
        }

        public static bool IsButtonDown(PlayerIndex playerIndex, Buttons button)
        {
            return currentGamePadState[(int)playerIndex].IsButtonDown(button);
        }

        public static bool IsButtonUp(PlayerIndex playerIndex, Buttons button)
        {
            return currentGamePadState[(int)playerIndex].IsButtonUp(button);
        }

        public static bool ButtonPressedOnAnyGamePad(Buttons button)
        {
            return ButtonPressed(PlayerIndex.One, button) ||
                ButtonPressed(PlayerIndex.Two, button) ||
                ButtonPressed(PlayerIndex.Three, button) ||
                ButtonPressed(PlayerIndex.Four, button);
        }

        public static bool ButtonReleasedOnAnyGamePad(Buttons button)
        {
            return ButtonReleased(PlayerIndex.One, button) ||
                ButtonReleased(PlayerIndex.Two, button) ||
                ButtonReleased(PlayerIndex.Three, button) ||
                ButtonReleased(PlayerIndex.Four, button);
        }
    }
    public enum MouseButton
    {
        Left,
        Right,
        Middle
    }

}
