using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;

namespace GameEngine.TextureGeneration
{
    public class BloodGenerator : ITextureGenerator
    {
        /// <summary>How much liquid are we throwing at the surface in ml (influences the size of the blood spatter)</summary>
        private int liquidAmount;
        public int LiquidAmount
        {
            get { return this.liquidAmount; }
            set { this.liquidAmount = (int)MathHelper.Clamp(value, 20f, 8192); }
        }

        /// <summary>The angle of the surface we have splattered blood upon. 0 is floor, PiOver2 is a wall.</summary>
        private float slope;
        public float Slope
        {
            get { return this.slope; }
            set { this.slope = MathHelper.Clamp(value, 0f, MathHelper.PiOver2); }
        }

        private Texture2D texture;
        private int realTextureSize;
        public int TextureSize { get { return this.realTextureSize; } }
        private GraphicsDevice gd;
        public GraphicsDevice GraphicsDevice { set { this.gd = value; } }

        private float[,] bloodTiles;
        private int canvasSize;

        /// <summary>
        /// Create a blood generator, which by default will create a circular blood spatter
        /// </summary>
        /// <param name="gd">The painting graphics device</param>
        public BloodGenerator(GraphicsDevice gd)
        {
            this.LiquidAmount = 1024;
            this.Slope = 0;
            this.gd = gd;
        }

        private void PaintDot(int xPos, int yPos, double amount, int dataWidth, Color[] data)
        {
            float redAmount = (float)(amount * 5d);

            Color redColor = new Color(1f, 0f, 0f, redAmount);

            data[(xPos * dataWidth) + yPos] = redColor;

            /*if (xPos == dataWidth - 1)
                data[(xPos * dataWidth) + yPos] = Color.Black;
            if (yPos == dataWidth - 1)0
                data[(xPos * dataWidth) + yPos] = Color.Black;*/
        }

        private void Iterate()
        {
            float seepageDiagonal = .6f;
            float seepageSlopeAdjustmentUp = 1f - (this.slope / MathHelper.PiOver2);
            float seepageSlopeAdjustmentDown = 1f + seepageSlopeAdjustmentUp;

            float currentSeepage, diagonalSeepage;
            int xPos, yPos;
            for (int j = 0; j < this.canvasSize - 2; j++)
            {
                for (int i = 0; i < this.canvasSize - 2; i++)
                {
                    xPos = j + 1;
                    yPos = i + 1;

                    if (bloodTiles[xPos, yPos] > 0)
                    {
                        currentSeepage = bloodTiles[xPos, yPos] / 10f;
                        diagonalSeepage = currentSeepage * seepageDiagonal;
                        // y-1
                        {
                            // x-1
                            bloodTiles[xPos - 1, yPos - 1] += (diagonalSeepage * seepageSlopeAdjustmentUp);
                            // x
                            bloodTiles[xPos, yPos - 1] += (currentSeepage * seepageSlopeAdjustmentUp);

                            // x+1
                            bloodTiles[xPos + 1, yPos - 1] += (diagonalSeepage * seepageSlopeAdjustmentUp);

                            bloodTiles[xPos, yPos] -= 3 * (diagonalSeepage * seepageSlopeAdjustmentUp);
                        }

                        // y
                        {
                            bloodTiles[xPos - 1, yPos] += currentSeepage;
                            bloodTiles[xPos + 1, yPos] += currentSeepage;
                            bloodTiles[xPos, yPos] -= 2 * currentSeepage;
                        }


                        // y+1
                        {
                            // x-1
                            bloodTiles[xPos - 1, yPos + 1] += (diagonalSeepage * seepageSlopeAdjustmentDown);
                            // x
                            bloodTiles[xPos, yPos + 1] += (currentSeepage * seepageSlopeAdjustmentDown);
                            // x+1
                            bloodTiles[xPos + 1, yPos + 1] += (diagonalSeepage * seepageSlopeAdjustmentDown);

                            bloodTiles[xPos, yPos] -= 3 * (diagonalSeepage * seepageSlopeAdjustmentDown);
                        }
                    }
                }
            }
        }

        #region ITextureGenerator Members

        public void Generate()
        {
            canvasSize = (this.LiquidAmount / 10);
            int halfWidth = canvasSize / 2;

            int spatterCount = this.liquidAmount / 30;
            int spatterAmount = (int)((float)this.LiquidAmount / (float)spatterCount);
            int spatterRadius = (int)((float)halfWidth * 0.60f);

            realTextureSize = 2;
            do
                realTextureSize *= 2;
            while (realTextureSize < canvasSize);
            // SurfaceFormat.Vector4 = alpha, blue, green, red
            this.texture = new Texture2D(gd, realTextureSize, realTextureSize, 0, TextureUsage.None, SurfaceFormat.Color);

            this.bloodTiles = new float[canvasSize, canvasSize];

            int xPos, yPos;
            double angle, currentRadius;
            for (int i = 0; i < spatterCount; i++)
            {
                angle = Util.Random.NextDouble() * (2 * MathHelper.Pi);
                currentRadius = ((double)i / (double)spatterCount) * (double)spatterRadius;

                xPos = (int)Math.Ceiling((currentRadius * Math.Cos(angle)));
                yPos = (int)Math.Ceiling((currentRadius * Math.Sin(angle)));

                // Add an appropriate amount of liquid (that is, the amount of blood in a single spatter)
                 bloodTiles[xPos + halfWidth, yPos + halfWidth] += spatterAmount;
            }


            for (int i = 0; i < spatterCount / 4; i++)
            {
                this.Iterate();
            }
        }

        public Texture2D GetTexture()
        {
            if (texture == null)
                this.Generate();

            float maxAmount = 0f;
            for (int i = 0; i < this.canvasSize; i++)
            {
                for (int j = 0; j < this.canvasSize; j++)
                {
                    maxAmount = (bloodTiles[i, j] > maxAmount) ? bloodTiles[i, j] : maxAmount;
                    if (maxAmount > 5f)
                        break;
                }
                if (maxAmount > 5f)
                    break;
            }

            Color[] data = new Color[realTextureSize * realTextureSize];
            for (int i = 0; i < data.Length; i++)
                data[i] = Color.TransparentBlack;

            int offset = (realTextureSize - canvasSize) / 2;

            double theDot, theValue;
            for (int i = 0; i < this.canvasSize; i++)
            {
                for (int j = 0; j < this.canvasSize; j++)
                {
                    theDot = (double)MathHelper.Clamp(this.bloodTiles[i, j], 0f, 5f);
                    theValue = theDot / (double)maxAmount;
                    PaintDot(i + offset, j + offset, theValue, this.realTextureSize, data);
                }
            }

            texture.SetData(data);
            texture.GenerateMipMaps(TextureFilter.Linear);
            return this.texture;
        }

        #endregion
    }
}
