using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Editor.Data;
using System.ComponentModel;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.Helpers;


namespace GameEngine.Xml
{
    public class PolygonNode : ObjectNode
    {
        private bool dirtied = true;
        private List<Vector2> points = new List<Vector2>();
        public List<Vector2> cachedNormals = new List<Vector2>();

        Vector2 topLeft = Vector2.Zero;
        Vector2 bottomRight = Vector2.Zero;
        
        [Browsable(false)]
        public List<Vector2> Points
        {
            get { return points; }
        }

        private bool shootThrough = false;
        [Category("Polygon")]
        [Description("Shoot through")]
        public bool ShootThrough
        {
            get { return this.shootThrough; }
            set
            {
                if (this.shootThrough == value)
                    return;

                this.shootThrough = value;
                Dirty();
            }
        }

        public PolygonNode(string name) : base(name)
        {

        }



        public void Dirty()
        {
            dirtied = true;
        }

        public override void Update(Microsoft.Xna.Framework.GameTime time, DrawInfo drawInfo)
        {
            if (dirtied)
            {
                ResolveDirty();
            }
            base.Update(time,drawInfo);
        }

        public virtual void ResolveDirty()
        {
            CalculateCollisionInfo();
            dirtied = false;
        }

        public void SetPoints(Vector2[] points)
        {
            dirtied = true;
            this.points.Clear();
            this.points.AddRange(points);
        }

        public void AddPoint(int insertAt, Vector2 point)
        {
            dirtied = true;
            points.Insert(insertAt, point);
        }

        public void RemovePoint(int removeAt)
        {
            dirtied = true;
            points.RemoveAt(removeAt);
        }

        public void ChangeWinding()
        {
            points.Reverse();
            dirtied = true;
        }

        public override void PopulateXml(System.Xml.XmlElement element, System.Xml.XmlDocument document)
        {
            base.PopulateXml(element, document);
            foreach (Vector2 point in points)
            {
                element.AppendChild(XmlHelper.CreateElement("Point", point, document));
            }

            element.SetAttribute("ShootThrough", this.shootThrough.ToString().ToLower());

        }
        public override void SetFromXml(System.Xml.XmlElement element)
        {
            base.SetFromXml(element);
            foreach (System.Xml.XmlElement cNode in element.ChildNodes)
            {
                switch (cNode.Name)
                {
                    case "Point":
                        Points.Add(XmlHelper.ParseVector2(cNode));
                        break;
                    default:
                        break;
                }
            }

            if (element.HasAttribute("ShootThrough"))
                this.ShootThrough = Convert.ToBoolean(element.Attributes["ShootThrough"].Value);
        }



        #region Collision stuff
        [Browsable(false)]
		public bool IsConcave { get; private set; }
        [Browsable(false)]
        public bool IsInsideOut { get; private set; }
        [Browsable(false)]
        public bool HasMixedAngles { get; private set; }

        public bool Intersects(BoundingCircle circle, Vector2 lastMove, out Vector2 resolveVector)
        {
            return IntersectsMixed(circle, lastMove, out resolveVector);
/*
            if (HasMixedAngles)
            {
                return IntersectsMixed(circle, out resolveVector);
            }
            else
            {
                if (IsConcave)
                {
                    return IntersectsConcave(circle, out resolveVector);
                }
                else
                {
                    return IntersectsConvex(circle, out resolveVector);
                }
            }*/
        }

        public bool Intersects(CollisionRay ray, out float collisionDist)
        {
            int lastIndex = points.Count - 1;
            float closestDist = float.MaxValue;
            bool collided = false;
            for (int i = 0; i < points.Count; i++)
            {
                float tmpDist;
                if (ray.Intersects(points[lastIndex], points[i] , out tmpDist)) 
                {
                    if (tmpDist < closestDist)
                        closestDist = tmpDist;
                    collided = true;
                }
                lastIndex = i;
            }

            collisionDist = closestDist;

            return collided;
        }

        public bool Intersects(CollisionRay ray, out float collisionDist, out Vector2 collisionNormal)
        {
            collisionNormal = Vector2.Zero;
            int lastIndex = points.Count - 1;
            float closestDist = float.MaxValue;
            bool collided = false;
            for (int i = 0; i < points.Count; i++)
            {
                float tmpDist;
                if (ray.Intersects(points[lastIndex], points[i], out tmpDist))
                {
                    if (tmpDist < closestDist)
                    {
                        closestDist = tmpDist;
                        collisionNormal = cachedNormals[lastIndex];
                    }
                    collided = true;
                }
                lastIndex = i;
            }

            collisionDist = closestDist;

            return collided;

        }


        private bool IntersectsConvex(BoundingCircle circle, out Vector2 resolve)
        {
            resolve = Vector2.Zero;

            // Early out from bounding box
            if (topLeft.X - circle.Center.X > circle.Radius)
                return false;
            if (circle.Center.X - bottomRight.X > circle.Radius)
                return false;
            if (topLeft.Y - circle.Center.Y > circle.Radius)
                return false;
            if (circle.Center.Y - bottomRight.Y > circle.Radius)
                return false;

            Vector2 testPoint = circle.Center;

            float outDist = float.MinValue;
            int outIndex = 0;

            for (int i = 0; i < points.Count; i++)
            {
                float dist = PointToEdgeDist(i, testPoint);
                if (dist > circle.Radius)
                {
                    return false;
                }
                else
                {
                    if (dist > outDist)
                    {
                        outDist = dist;
                        outIndex = i;
                    }
                }

            }

            resolve = cachedNormals[outIndex] * (circle.Radius - outDist);

            return true;
        }

        private bool IntersectsConcave(BoundingCircle circle, out Vector2 resolve)
        {
            Vector2 testPoint = circle.Center;

            resolve = Vector2.Zero;

            int collisionCount = 0;

            for (int i = 0; i < points.Count; i++)
            {
                float dist = PointToEdgeDist(i, testPoint + resolve);
                if (dist < circle.Radius)
                {
                    Vector2 normal = cachedNormals[i];
                    Vector2 newResolve = normal * (circle.Radius - dist);
                    /*
                    if(Vector2.Dot(resolve, newResolve) < -0.00001f) 
                    {
                        Vector2 oldEdgeDirection = new Vector2(-resolve.Y, resolve.X);
                        newResolve = oldEdgeDirection * newResolve.LengthSquared() / Vector2.Dot(newResolve, oldEdgeDirection);
                    }
                    */

                    resolve += newResolve;

                    collisionCount++;
                }
            }

            return (collisionCount > 0);
        }


        private float PointToEdgeDist(int edgeIndex, Vector2 point)
        {
            Vector2 normal = cachedNormals[edgeIndex];

            Vector2 startToCenter = point - points[edgeIndex];

            return Vector2.Dot(normal, startToCenter);
        }

        private bool IntersectsMixed(BoundingCircle circle, Vector2 lastMove, out Vector2 resolve)
        {

            Vector2 testPoint = circle.Center;

            // Default to no resolve vector
            resolve = Vector2.Zero;

            if (HasMixedAngles && IsInsideOut)
            {
                if (!ContainsPoint(testPoint))
                {
                    return true;
                }
            }
            else
            {
                // Early out from bounding box
                if (topLeft.X - circle.Center.X > circle.Radius)
                    return false;
                if (circle.Center.X - bottomRight.X > circle.Radius)
                    return false;
                if (topLeft.Y - circle.Center.Y > circle.Radius)
                    return false;
                if (circle.Center.Y - bottomRight.Y > circle.Radius)
                    return false;

                if (HasMixedAngles && ContainsPoint(testPoint))
                {
                    // Shoot a ray from last position to the new one:
                    CollisionRay ray = new CollisionRay(circle.Center - lastMove, lastMove, true);

                    float colDist;
                    Vector2 colNormal;
                    // Find the place where the ray intersects with the polygon and go the length of the radius out
                    // from this point to get a position where the collision is resolved
                    if (Intersects(ray, out colDist, out colNormal))
                    {
                        if (colDist > 100)
                        {
                        }
                        Vector2 resolvedPoint = ray.Position + ray.Direction * colDist + colNormal * circle.Radius;
                        resolve = resolvedPoint - circle.Center;
                    }

                    return true;
                }
            }

            // Check if circle intersects any edge line segments:
            int collisionCount = 0;
            int startIndex = points.Count - 1;
            for (int i = 0; i < points.Count; i++)
            {
                Vector2 newResolve = Vector2.Zero;
                bool collision = true;

                Vector2 currentSegment = points[i] - points[startIndex];
                float segmentLength = currentSegment.Length();
                currentSegment /= segmentLength;

                Vector2 startToCenter = circle.Center + resolve - points[startIndex];
                float d = Vector2.Dot(startToCenter, currentSegment);
                Vector2 normal = new Vector2(-currentSegment.Y, currentSegment.X);
                float dist = Vector2.Dot(normal, startToCenter);

                if (dist < 0)
                {
                    collision = false;
                }
                else if (d < 0)
                {
                    // circle "before" the line segment, check distance from center to startpoint
                    if (startToCenter.LengthSquared() < circle.Radius * circle.Radius)
                    {
                        float centerDist = startToCenter.Length();
                        newResolve = startToCenter / centerDist * (circle.Radius - centerDist);
                    }
                    else
                        collision = false;
                }
                else if (d < segmentLength)
                {
                    if (dist > 0 && dist < circle.Radius)
                    {
                        newResolve = normal * (circle.Radius - dist);
                    }
                    else
                        collision = false;
                }
                else
                {
                    // The case where the center is "after" the line segment will be handled by next iteration
                    collision = false;
                }

                if (collision)
                {
                    /*
                    if (Vector2.Dot(resolve, newResolve) < -0.00001f)
                    {
                        Vector2 oldEdgeDirection = new Vector2(-resolve.Y, resolve.X);
                        resolve += oldEdgeDirection * newResolve.LengthSquared() / Vector2.Dot(newResolve, oldEdgeDirection);
                    }
                    else
                    {
                        resolve += newResolve;
                    }
                    */
                    resolve += newResolve;
                    collisionCount++;
                }

                startIndex = i;
            }


            return collisionCount > 0;
        }

        public bool LineSegmentCircleIntersection(BoundingCircle circle, Vector2 startPoint, Vector2 endPoint, out Vector2 resolve)
        {
            Vector2 startToEnd = endPoint - startPoint;
            float segmentLength = startToEnd.Length();
            startToEnd /= segmentLength;

            Vector2 startToCenter = circle.Center - startPoint;
            float d = Vector2.Dot(startToCenter, startToEnd);

            if (d < 0)
            {
                if (startToCenter.LengthSquared() < circle.Radius * circle.Radius)
                {
                    float centerDist = startToCenter.Length();
                    resolve = startToCenter / centerDist * (circle.Radius - centerDist);
                    return true;
                }
            }
            else if (d <= segmentLength)
            {
                Vector2 normal = new Vector2(-startToEnd.Y, startToEnd.X);

                float dist = Vector2.Dot(normal, startToCenter);
                if (dist < circle.Radius)
                {
                    resolve = normal * (circle.Radius - dist);
                    return true;
                }
            }
            else
            {
                Vector2 endToCenter = circle.Center - endPoint;
                if (endToCenter.LengthSquared() < circle.Radius * circle.Radius)
                {
                    float centerDist = endToCenter.Length();
                    resolve = endToCenter / centerDist * (circle.Radius - centerDist);
                    return true;
                }
            }

            resolve = Vector2.Zero;
            return false;
        }



        public bool ContainsPoint(Vector2 testPoint)
        {
            int intersectCount = 0;
            int lastIndex = points.Count - 1;
            for (int i = 0; i < points.Count; i++)
            {
                if (RaySegmentTest(testPoint, points[lastIndex], points[i]))
                    intersectCount++;
                lastIndex = i;
            }

            return (!((intersectCount % 2) == 0));
        }

        // Tests intersection between the ray stating at rayStart in the direction
        // of the positive X-axis and the given line section.
        private bool RaySegmentTest(Vector2 rayStart, Vector2 segStart, Vector2 segEnd)
        {
            // If we're above or below the ray, return false
            if ((segStart.Y > rayStart.Y && segEnd.Y > rayStart.Y) || (segStart.Y < rayStart.Y && segEnd.Y < rayStart.Y))
                return false;

            if (segStart.X < rayStart.X)
            {
                if (segEnd.X < rayStart.X)
                {
                    return false;
                }
            }
            else
            {
                if (segEnd.X >= rayStart.X)
                {
                    // start after, end after
                    return true;
                }
            }

            // start after, end before or start before, end after
            Vector2 lineSeg = segEnd - segStart;
            Vector2 normal = new Vector2(-lineSeg.Y, lineSeg.X);
            return lineSeg.Y > 0 ? Vector2.Dot(normal, rayStart - segStart) >= 0 : Vector2.Dot(normal, rayStart - segStart) <= 0;

        }


        private void CalculateCollisionInfo()
        {
            // Reset data
            int leftMostPointIndex = points.Count - 1;
            topLeft = bottomRight = points[leftMostPointIndex];
            cachedNormals.Clear();


            int concaveCount = 0;
            Vector2 lastSegment = points[0] - points[leftMostPointIndex];

            for (int i = 0; i < points.Count; i++)
            {

                // Find topLeft and bottomRight as well as leftmost point
                if (points[i].X < topLeft.X)
                {
                    topLeft.X = points[i].X;
                    leftMostPointIndex = i;
                }
                topLeft.Y = Math.Min(topLeft.Y, points[i].Y);
                bottomRight.X = Math.Max(bottomRight.X, points[i].X);
                bottomRight.Y = Math.Max(bottomRight.Y, points[i].Y);

                
                // Check for concave / convex angles
                int nextIndex = i + 1;
                if (nextIndex == points.Count)
                    nextIndex = 0;

                Vector2 currentSegment = points[nextIndex] - points[i];
                Vector2 normal = new Vector2(-currentSegment.Y, currentSegment.X);
                normal.Normalize();

                if (Vector2.Dot(normal, lastSegment) < 0)
                {
                    concaveCount++;
                }

                // Cache normals for later use:
                cachedNormals.Add(normal);

                lastSegment = currentSegment;
            }

            if (concaveCount == 0)
            {
                IsConcave = false;
                IsInsideOut = false;
                HasMixedAngles = false;
            }
            else if (concaveCount == points.Count)
            {
                IsConcave = true;
                IsInsideOut = true;
                HasMixedAngles = false;
            }
            else
            {
                IsConcave = true;
                HasMixedAngles = true;

                // Average between the two edges sourrounding the leftmost point
                int beforeIndex = leftMostPointIndex == 0 ? points.Count - 1 : leftMostPointIndex - 1;
                int afterIndex = leftMostPointIndex == (points.Count - 1) ? 0 : leftMostPointIndex + 1;

                Vector2 beforeVector = points[leftMostPointIndex] - points[beforeIndex];
                Vector2 afterVector = points[afterIndex] - points[leftMostPointIndex];

                Vector2 avgVector = beforeVector + afterVector;

                IsInsideOut = (Vector2.Dot(Vector2.UnitY, avgVector) < 0.00001f);
            }
        }

	    #endregion
    }
}
