using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Microsoft.Xna.Framework;
using System.ComponentModel;

namespace GameEngine.Xml
{

    public class SceneNode 
    {
        protected string name;
        private SceneNode parent;
        private NodeList childNodes;
        private string include = null;
        private object tag;

        [Browsable(false)]
        public SceneNode Parent { get { return parent; } set { parent = value; } }

        [Browsable(false)]
        public NodeList ChildNodes
        {
            get { return childNodes; }
        }

        public SceneNode(string name)
        {
            this.Name = name;
            childNodes = new NodeList(this);
        }

        [Category("Node")]
        public virtual string Name { get { return name; } set { name = value; } }
        [Category("Node")]
        public virtual string Path { get { return parent != null ? parent.Path + "." + Name : Name; } }


        public virtual void PopulateXml(XmlElement element, XmlDocument document)
        {
            if (include != null)
                element.SetAttribute("Include", include);
            element.SetAttribute("Name", name);
        }

        public virtual void SetFromXml(XmlElement element)
        {
            if(element.HasAttribute("Name"))
                Name = element.Attributes["Name"].Value;

            if (element.HasAttribute("Include") && !string.IsNullOrEmpty(element.Attributes["Include"].Value.Trim()))
                Include = element.Attributes["Include"].Value;
        }

        [Browsable(false)]
        public virtual string Include
        {
            get { return include; }
            set { include = value; }
        }

        [Browsable(false)]
        public object Tag
        {
            get { return tag; }
            set { tag = value; }
        }


    }

    public class NodeList : IList<SceneNode>
    {
        private List<SceneNode> nodes = new List<SceneNode>();
        public SceneNode Owner { get; private set; }

        public NodeTraverser NodesBreadthFirst { get { return new NodeTraverser(this, true); } }
        public NodeTraverser NodesDepthFirst { get { return new NodeTraverser(this, false); } }

        public NodeList(SceneNode owner)
        {
            this.Owner = owner;
        }

        #region IList<SceneNode> Members

        public int IndexOf(SceneNode item)
        {
            return nodes.IndexOf(item);
        }

        public void Insert(int index, SceneNode item)
        {
            if (item == null)
                return;

            item.Parent = Owner;

            nodes.Insert(index, item);
        }

        public void RemoveAt(int index)
        {
            nodes.RemoveAt(index);
        }

        public SceneNode this[int index]
        {
            get
            {
                return nodes[index];
            }
            set
            {
                nodes[index] = value;
            }
        }

        #endregion

        #region ICollection<SceneNode> Members

        public void Add(SceneNode item)
        {
            if (item == null)
                return;

            item.Parent = Owner;

            nodes.Add(item);
        }

        public void AddRange(IEnumerable<SceneNode> items)
        {
            foreach (var item in items)
            {
                this.Add(item);
            }
        }

        public void Clear()
        {
            nodes.Clear();
        }

        public bool Contains(SceneNode item)
        {
            return nodes.Contains(item);
        }

        public void CopyTo(SceneNode[] array, int arrayIndex)
        {
            nodes.CopyTo(array, arrayIndex);
        }

        public int Count
        {
            get { return nodes.Count; }
        }

        public bool IsReadOnly
        {
            get { return false; }
        }

        public bool Remove(SceneNode item)
        {
            return nodes.Remove(item);
        }

        #endregion

        #region IEnumerable<SceneNode> Members

        public IEnumerator<SceneNode> GetEnumerator()
        {
            return nodes.GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return nodes.GetEnumerator();
        }

        public SceneNode this[string index] 
        { 
            get {
                foreach (SceneNode node in nodes)
                {
                    if (node.Name == index)
                        return node;
                }
                throw new ArgumentException("Could not find item in collection");
            } 
        }

        public bool Contains(string key)
        {
            foreach (SceneNode node in nodes)
            {
                if (node.Name == key)
                    return true;
            }
            return false;
        }

        #endregion



    }


}
