﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.ScreenSystem;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using GameEngine.Cameras;
using GamePrototype.GameLogic;
using GameEngine.Physics;
using GameEngine.Helpers;

namespace GamePrototype.GUI.Screens
{
    public class JanDebugFraHelvede : GameScreen
    {
        private OrthoOrbitCamera camera;
        private Model sandstone;

        private BasicEffect basicEffect;
        private VertexDeclaration vertexDeclaration;
        private VertexPositionColor[] pointList;
        private VertexBuffer vertexBuffer;
        private int activeCount = 0;

        private const int PARTICLE_COUNT = 100;
        private const int PARTICLE_SIZE = 5;
        private Color traceColor = Color.White;
        private Color bloodColor = Color.Red;
        private Color fragmentColor = Color.Yellow;

        private TraceEmitter traceEmitter;
        private FragmentEmitter fragmentEmitter;
        private BloodEmitter bloodEmitter;
        Quad q;
        public JanDebugFraHelvede(string id, ScreenController controller, Game game)
            : base(id, controller, game)
        {
            traceEmitter = new TraceEmitter(PARTICLE_COUNT);
            fragmentEmitter = new FragmentEmitter(PARTICLE_COUNT);
            bloodEmitter = new BloodEmitter(PARTICLE_COUNT);
        }

        public override void LoadContent(GraphicsDeviceManager graphics)
        {
            graphics.PreferredBackBufferWidth = 1024;
            graphics.PreferredBackBufferHeight = 768;
            graphics.ApplyChanges();
            
            sandstone = content.Load<Model>("models/Sandstone/sandstone");

            camera = new OrthoOrbitCamera(MathHelper.PiOver4, new Vector3(600, 300, 600), new Vector3(0, 0, 0), Vector3.Up, 
                graphics.GraphicsDevice.Viewport, 1f, 100000f, CameraMode.Perspective);

            InitPointDrawer(graphics.GraphicsDevice);
        }

        public override void Initialize(GraphicsDevice graphics)
        {
                           
        }

        private void InitPointDrawer(GraphicsDevice graphics)
        {
            vertexDeclaration = new VertexDeclaration(graphics, VertexPositionColor.VertexElements);
            basicEffect = new BasicEffect(graphics, null);
            basicEffect.Texture = content.Load<Texture2D>("enginefx\\trace");
            basicEffect.TextureEnabled = true;
            pointList = new VertexPositionColor[PARTICLE_COUNT * 3];
            q = Quad.GenerateXZ(graphics, 100, 0, Vector2.Zero, Vector2.Zero, Vector4.One);
            SetupPoints();

            vertexBuffer = new VertexBuffer(graphics, VertexPositionColor.SizeInBytes * (pointList.Length), BufferUsage.None);
            vertexBuffer.SetData<VertexPositionColor>(pointList);
        }

        public override void Update(GameTime gameTime)
        {
            float zoomFactor = 10f;
            float orbitSpeed = 0.04f;
            
            if (Keyboard.GetState().IsKeyDown(Keys.Escape)) { Game.Exit(); }

            if (Keyboard.GetState().IsKeyDown(Keys.Q)) { camera.Zoom(zoomFactor); }
            if (Keyboard.GetState().IsKeyDown(Keys.E)) { camera.Zoom(-zoomFactor); }
            if (Keyboard.GetState().IsKeyDown(Keys.D)) { camera.Orbit(orbitSpeed, 0); }
            if (Keyboard.GetState().IsKeyDown(Keys.A)) { camera.Orbit(-orbitSpeed, 0); }
            if (Keyboard.GetState().IsKeyDown(Keys.S)) { camera.Orbit(0, orbitSpeed); }
            if (Keyboard.GetState().IsKeyDown(Keys.W)) { camera.Orbit(0, -orbitSpeed); }

            if (InputState.KeyPressed(Keys.T))
            {
                CollisionRay ray = new CollisionRay(new Vector2(0, 0), new Vector2(0, 1), true);
                float height = 70;
                
                traceEmitter.Trace(ray, height, 20);  
            }

            if (InputState.IsKeyDown(Keys.F))
            {
                int fragmentCount = 5;
                float height = 70;
                Vector2 impactPosition = new Vector2(0, 0);
                Vector2 fromDirection = new Vector2(-5, -5);
                Vector2 impactNormal = new Vector2(0, 1);

                fragmentEmitter.DoFragments(impactPosition, impactNormal, fromDirection, height, fragmentCount);
            }

            if (InputState.KeyPressed(Keys.B))
            {
                int fragmentCount = 5;
                float height = 70;
                Vector2 impactPosition = new Vector2(0, 0);
                Vector2 fromDirection = new Vector2(-5, -5);
                Vector2 impactNormal = new Vector2(0, 1);
                bloodEmitter.DoFragments(impactPosition, fromDirection, height, fragmentCount);
            }

            traceEmitter.Update(gameTime);
            fragmentEmitter.Update(gameTime);
            bloodEmitter.Update(gameTime);

            camera.UpdateCamera(gameTime);
        }

        public override void Draw(GameTime gameTime, GraphicsDevice graphics, SpriteBatch spriteBatch, SpriteFont spriteFont)
        {
            graphics.Clear(Color.Black);

            SetupPoints();
            graphics.RenderState.CullMode = CullMode.None;
            if (activeCount > 0)
            {
                basicEffect.World = Matrix.CreateScale(10) ;
                basicEffect.Projection = camera.Projection;
                basicEffect.View = camera.View;

                q.Draw(graphics, basicEffect);

            }

            DrawBoxes();

            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.SaveState);
            spriteBatch.DrawString(spriteFont, "HEJ!", new Vector2(20, 20), Color.Red);
            spriteBatch.End();
        }

        private void DrawModel(Model m, Matrix world, Matrix view, Matrix projection)
        {
            foreach (ModelMesh mesh in m.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.World = world;
                    effect.View = view;
                    effect.Projection = projection;
                    effect.EnableDefaultLighting();
                }

                mesh.Draw();
            }
        }

        private void DrawBoxes()
        {
            float scale = 0.1f;
            float minPos = -1000;
            float maxPos = 1000;

            DrawModel(sandstone, Matrix.CreateTranslation(new Vector3(minPos, minPos, 0)) * Matrix.CreateScale(scale), camera.View, camera.Projection);
            DrawModel(sandstone, Matrix.CreateTranslation(new Vector3(maxPos, minPos, 0)) * Matrix.CreateScale(scale), camera.View, camera.Projection);
            DrawModel(sandstone, Matrix.CreateTranslation(new Vector3(minPos, maxPos, 0)) * Matrix.CreateScale(scale), camera.View, camera.Projection);
            DrawModel(sandstone, Matrix.CreateTranslation(new Vector3(maxPos, maxPos, 0)) * Matrix.CreateScale(scale), camera.View, camera.Projection);
        }
        
        private void SetupPoints()
        {
            activeCount = 0;

            for (int i = activeCount; i < traceEmitter.ActiveParticles.Count; i++)
            {
                pointList[i] = new VertexPositionColor(traceEmitter.ActiveParticles[i].Position, traceColor);
                activeCount++;
            }

            for (int i = activeCount; i < fragmentEmitter.ActiveParticles.Count; i++)
            {
                pointList[i] = new VertexPositionColor(fragmentEmitter.ActiveParticles[i].Position, fragmentColor);
                activeCount++;
            }

            for (int i = activeCount; i < bloodEmitter.ActiveParticles.Count; i++)
            {
                pointList[i] = new VertexPositionColor(bloodEmitter.ActiveParticles[i].Position, bloodColor);
                activeCount++;
            }
        }
    }
}
