﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.ScreenSystem;
using GameEngine;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Input;

namespace GamePrototype.GameLogic
{
    public class Weapon
    {
        private const float WEAPON_SOUND_VOLUME = 0.8f;
        private const int WEAPON_EMPTY_DELAY = 500; // Prevent empty sound from having same frequency as rate of fire.

        public WeaponType Type { get; set; }
        public AmmoType AmmoType { get; set; }
        public int Damage { get; set; }
        public int RateOfFire { get; set; }
        public int ReloadTime { get; set; }
        public int ClipSize { get; set; }
        public int AmmoInClip { get; set; }
        public float AllowedDeviation { get; set; }
        public int Penetration { get; set; }
        public int ProjectileCount { get; set; }
        public string FireSound { get; set; }
        public string FireDrySound { get; set; }
        public string ReloadSound { get; set; }
        public bool SemiAutomatic { get; set; }
        public bool IsReloading { get; private set; }
        
        private bool hasPlayedReloadSound;
        private float msSinceLastFire;
        private float msSinceStartingReload;
        
        public Weapon(WeaponType type)
        {
            Type = type;
        }

        public void PrepareWeapon()
        {
            
            AmmoInClip = ClipSize; // Always start with a full clip
            msSinceLastFire = ReloadTime; // Makes it possible to fire immediately 
            msSinceStartingReload = 0;
            IsReloading = false;
            hasPlayedReloadSound = false;
        }

        public float GetReloadProgress()
        {
            return (float)msSinceStartingReload / (float)ReloadTime;
        }

        public void Update(GameTime gameTime, Player player)
        {
            msSinceLastFire += gameTime.ElapsedMillis();

            if (AmmoInClip < 1 && player.GetAmmoCount(this.AmmoType) > 0)
            {
                 TryFire(player, false); // Don't ask
            }

            if (IsReloading)
            {
                msSinceStartingReload += gameTime.ElapsedMillis();
                                
                if (!hasPlayedReloadSound)
                {
                    if (msSinceStartingReload > ReloadTime / 2)
                    {
                        AudioManager.PlaySoundEffect(ReloadSound, WEAPON_SOUND_VOLUME);
                        hasPlayedReloadSound = true;        
                    }
                }

                if (msSinceStartingReload >= ReloadTime) // Done reloading?
                {
                    AmmoInClip = ClipSize;
                    IsReloading = false;
                }                
            }                        
        }

        public bool TryFire(Player player, bool justTriggered)
        {
            if (msSinceLastFire > RateOfFire) // Check rate of fire, so we can't fire too fast 
            {
                if (!IsReloading)
                {
                    if (AmmoInClip > 0) // Any bullets in current clip?  
                    {
                        if (!SemiAutomatic || justTriggered)
                        {
                            AmmoInClip--;
                            msSinceLastFire = 0;
                            AudioManager.PlaySoundEffect(FireSound, WEAPON_SOUND_VOLUME);
                            return true;
                        }
                        return false;
                                                
                    }
                    else // Clip is empty, try reloading
                    {
                        if (IsReloading)
                        {
                            return false; // Already reloading, nothing to do
                        }
                        else // Play fire empty sound and try to reload
                        {
                            if (player.GetAmmoCount(this.AmmoType) > 0)
                            {
                                AudioManager.PlaySoundEffect(FireDrySound, WEAPON_SOUND_VOLUME);    
                            }
                            else
                            {
                                AudioManager.PlaySoundEffect(ReloadSound, WEAPON_SOUND_VOLUME);
                            }

                            msSinceLastFire = -WEAPON_EMPTY_DELAY;

                            if (player.GetAmmoCount(this.AmmoType) > 0)
                            {
                                player.DecreaseAmmoCount(this.AmmoType);
                                msSinceStartingReload = 0;
                                hasPlayedReloadSound = false;
                                IsReloading = true;
                                return false;
                            }
                        }
                    }    
                }
            }
                        
            return false;
        }
        
        public void Reload(Player player)
        {
            if (!IsReloading)
            {
                if (player.GetAmmoCount(this.AmmoType) > 0)
                {
                    player.DecreaseAmmoCount(this.AmmoType);
                    msSinceStartingReload = 0;
                    IsReloading = true;
                    hasPlayedReloadSound = false;
                }
            }
        }

        public void AbortReload()
        {
            IsReloading = false;
            msSinceStartingReload = 0;
        }
    }
}
