﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using GameEngine.Helpers;
using GameEngine.Xml;
using GameEngine.LevelGeneration;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using GameEngine;
using Microsoft.Xna.Framework.Input;

namespace GamePrototype.GameLogic
{
    public class ZombieController
    {
        public int difficulty;
        public int[] HITPOINTS = new int[] { 150, 225, 225 };
        public int[] DAMAGE = new int[] { 6, 10, 15 };
        private int[] MIN_SPEED = new int[] { 200, 220, 250 };
        private int[] MAX_SPEED = new int[] { 250, 310, 330 };

        private Enemy[] enemyPool;
        private Player[] players;
        private SceneNode[] lastPlayerParent;
        private CollisionSimulator collisionSimulator;

        private LevelTile[,] levelTiles;
        private Point levelOffset;
        private ContentManager content;
        private GraphicsDevice graphics;

        private List<LevelTileNode> spawnTiles = new List<LevelTileNode>();
        private LevelTileNode startTile;

        int initialEnemiesPerTile = 10;
        int maxEnemies = 100;
        int spawnedEnemies = 0;

        int ticsBetweenSpawns = 20;
        int toNextSpawn = 0;

        public ZombieController(int difficulty, Player[] players, CollisionSimulator collisionSimulator, LevelTile[,] levelTiles, Point levelOffset, ContentManager content, GraphicsDevice graphics, GameWrapper wrapper)
        {
            this.difficulty = difficulty;
            this.players = players;
            this.collisionSimulator = collisionSimulator;

            this.levelTiles = levelTiles;
            this.levelOffset = levelOffset;
            this.content = content;
            this.graphics = graphics;

            enemyPool = new Enemy[maxEnemies];
            for (int i = 0; i < maxEnemies; i++)
            {
                ZombieBear zombie = new ZombieBear("zomb" + i, DAMAGE[difficulty - 1], HITPOINTS[difficulty - 1], GetMovementSpeed(difficulty),
                    collisionSimulator, content, wrapper);

                zombie.ChildNodes.LoadRecursive(content, graphics);
                //zombie.SetSuper(1f + (float)Util.Random.NextDouble() * 0.5f);
                zombie.Model.StartAnimation("TwoSteps", true);
                enemyPool[i] = zombie;
            }

        }

        public void Initialize()
        {


            for (int i = 0; i < levelTiles.GetLength(0); i++)
            {
                for (int j = 0; j < levelTiles.GetLength(1); j++)
                {
                    if (levelTiles[i, j] != null)
                    {
                        (levelTiles[i, j].SceneNode as LevelTileNode).SpawnCount = initialEnemiesPerTile;
                    }
                }
            }

            lastPlayerParent = new SceneNode[players.Length];
            for (int i = 0; i < players.Length; i++)
            {
                lastPlayerParent[i] = players[i].Parent;
                LevelTileNode playerTileNode = players[i].Parent as LevelTileNode;
                AddInterestTiles(new Point(playerTileNode.Tile.X + levelOffset.X, playerTileNode.Tile.Y + levelOffset.Y));
                startTile = playerTileNode;
            }

            int initialEnemies = initialEnemiesPerTile * spawnTiles.Count;

            for (int i = 0; i < initialEnemies; i++)
            {
                SpawnEnemyInRandomTile();
            }

            toNextSpawn = ticsBetweenSpawns;
        }

        float sinceLastTic = 0;
        float ticTime = 0.1f;

        public void Update(GameTime gameTime)
        {
            sinceLastTic += gameTime.ElapsedSeconds();
            if (sinceLastTic > ticTime)
            {
                DoTic();
                sinceLastTic = 0;
            }
        }

        private void DoTic()
        {
            // Check if we should add spawncontrollers for newly discovered rooms
            CheckPlayerNodeChange();

            toNextSpawn--;
            if (toNextSpawn <= 0 && spawnedEnemies < maxEnemies)
            {
                toNextSpawn = ticsBetweenSpawns;
                SpawnEnemyInRandomTile();
            }

            // Enemy control
            for (int i = enemyPool.Length - 1; i >= 0; i--)
            {
                Enemy enemy = enemyPool[i];

                if (! enemy.IsInUse || enemy.Dead || enemy.DisableTic)
                    continue;

                // TODO:
                //if(enemy.HasFinishedDying) {
                // DespawnEnemy(enemy);
                // continue;
                //}

                enemy.TicCounter++;

                // If in same tile as target, update waypoint every tic
                if (enemy.CurrentTarget.Parent == enemy.Parent)
                {
                    enemy.SetNextWaypoint(enemy.CurrentTarget.Position);
                    enemy.TicCounter = 0;
                }
                else
                {
                    if (enemy.TicCounter >= 20)
                    {
                        enemy.TicCounter = 0;

                        // TODO: Detect if we can get to player by moving to the tile he's in:
                        //int xDiff = (enemy.CurrentTarget.Parent as LevelTileNode).Tile.X - (enemy.Parent as LevelTileNode).Tile.X;
                        //int yDiff = (enemy.CurrentTarget.Parent as LevelTileNode).Tile.Y - (enemy.Parent as LevelTileNode).Tile.Y;

                        // Shout out rays in random directions and select the furthest target we hit
                        float maxDist = 0;
                        CollisionRay testRay = new CollisionRay(new Vector2(enemy.Position.X, enemy.Position.Z), Vector2.UnitX);
                        Vector2 target = Vector2.Zero;
                        for (int j = 0; j < 4; j++)
                        {
                            float angle = (float)(Util.Random.NextDouble() * Math.PI * 2);
                            testRay.SetAngleFromXAxis(angle);
                            
                            float testDist = collisionSimulator.GetClosestPolygonCollisionDistance(testRay, (enemy.Parent as LevelTileNode).Tile);
                            if (testDist < 1)
                            {
                            }
                            if (testDist > maxDist)
                            {
                                target = testRay.Position + testRay.Direction * maxDist;
                                maxDist = testDist;
                            }
                        }
                        enemy.SetNextWaypoint(new Vector3(target.X, 0, target.Y));
                    }
                }
            }
        }

        int INTERESTDISTANCE = 2;

        private bool IsCloseToPlayer(LevelTileNode tileNode)
        {
            for (int i = 0; i < players.Length; i++)
            {
                LevelTileNode playerTileNode = players[i].Parent as LevelTileNode;

                if (Math.Abs(tileNode.Tile.X - playerTileNode.Tile.X) <= INTERESTDISTANCE && Math.Abs(tileNode.Tile.Y - playerTileNode.Tile.Y) <= INTERESTDISTANCE)
                    return true;
            }

            return false;
        }

        private List<LevelTileNode> newlyAddedTiles = new List<LevelTileNode>();

        public int AddInterestTiles(Point startCoords) 
        {
            int x = Math.Max(startCoords.X - INTERESTDISTANCE, 0);
            int ex = Math.Min(startCoords.X + INTERESTDISTANCE, levelTiles.GetLength(0) - 1);

            int y = Math.Max(startCoords.Y - INTERESTDISTANCE, 0);
            int ey = Math.Min(startCoords.Y + INTERESTDISTANCE, levelTiles.GetLength(1) - 1);

            int addedTiles = 0;

            Point tileCoords;
            for (tileCoords.X = x; tileCoords.X <= ex; tileCoords.X++)
            {
                for (tileCoords.Y = y; tileCoords.Y <= ey; tileCoords.Y++)
                {
                    LevelTile tile = levelTiles[tileCoords.X, tileCoords.Y];
                    if (tile != null)
                    {
                        LevelTileNode tileNode = tile.SceneNode as LevelTileNode;
                        if (!tileNode.IsEmpty)
                        {
                            if (!spawnTiles.Contains(tileNode) && !ContainsPlayer(tileNode) && tileNode != startTile)
                            {
                                spawnTiles.Add(tileNode);
                                newlyAddedTiles.Add(tileNode);
                                addedTiles++;
                            }
                        }
                    }
                }
            }

            return addedTiles;
        }

        public int RemoveUninterstingTiles()
        {
            int enemiesRemoved = 0;
            for (int i = spawnTiles.Count - 1; i >= 0; i--)
            {
                if (!IsCloseToPlayer(spawnTiles[i]))
                {
                    for (int j = spawnTiles[i].DynamicCount - 1; j >= 0; j--)
                    {
                        if (spawnTiles[i][j] is Enemy)
                        {
                            DespawnEnemy(spawnTiles[i][j] as Enemy);
                            enemiesRemoved++;
                        }
                    }
                    spawnTiles[i].SpawnCount = initialEnemiesPerTile;
                    spawnTiles.RemoveAt(i);
                }
            }

            for (int i = 0; i < players.Length; i++)
            {
                spawnTiles.Remove(players[i].Parent as LevelTileNode);
            }

            return enemiesRemoved;
        }

        private bool ContainsPlayer(LevelTileNode node)
        {
            for (int i = 0; i < node.DynamicCount; i++)
            {
                if (node[i] is Player)
                    return true;
            }

            return false;
        }


        private void CheckPlayerNodeChange()
        {
            bool updateInterests = false;
            for (int i = 0; i < players.Length; i++)
            {
                if (players[i].Parent != lastPlayerParent[i])
                {
                    lastPlayerParent[i] = players[i].Parent;
                    updateInterests = true;
                }
            }

            if (updateInterests)
            {
                // Remove tiles and enemies that are no longer close to the players
                int enemiesRemoved = RemoveUninterstingTiles();
                if (enemiesRemoved > 0)
                {
                    enemiesRemoved++;
                }

                newlyAddedTiles.Clear();

                int addedTiles = 0;

                for (int i = 0; i < players.Length; i++)
                {
                    // If the player has moved check the surrounding tiles
                    LevelTileNode playerTile = players[i].Parent as LevelTileNode;
                    addedTiles += AddInterestTiles(new Point(playerTile.Tile.X + levelOffset.X, playerTile.Tile.Y + levelOffset.Y));
                }
                int newEmemyCount = addedTiles * initialEnemiesPerTile;

                for (int i = 0; i < newEmemyCount; i++)
                {
                    // Spawn the removed enemies in random places in the new tiles
                    SpawnEnemyInTile(newlyAddedTiles[Util.Random.Next(newlyAddedTiles.Count)]);
                }
            }
        }

        public Enemy GetNextUnusedEnemy()
        {
            for (int i = 0; i < enemyPool.Length; i++)
            {
                if (!enemyPool[i].IsInUse)
                    return enemyPool[i];
            }

            return null;
        }

        public void SpawnEnemyInRandomTile()
        {
            SpawnEnemyInTile(spawnTiles[Util.Random.Next(spawnTiles.Count)]);
        }

        public void SpawnEnemyInTile(LevelTileNode tileNode)
        {
            if (tileNode.SpawnCount <= 0)
                return;

            Enemy enemy = GetNextUnusedEnemy();
            if (enemy != null)
            {
                SpawnPointNode pp = tileNode.GetAvailableSpawnPoint(SpawnPointNode.SpawnType.BigZombie);

                SpawnPointNode spawnPoint = pp == null ? tileNode.GetAvailableSpawnPoint(SpawnPointNode.SpawnType.Zombie) : pp;
                if (spawnPoint == null)
                    return;
                
                enemy.IsInUse = true;

                tileNode.AddDynamicNode(enemy);
                tileNode.SpawnCount--;
                enemy.hitpoints = HITPOINTS[difficulty - 1];
                enemy.CurrentTarget = players[Util.Random.Next(players.Length)];
                enemy.SetNextWaypoint(enemy.CurrentTarget.Position);
                enemy.TicCounter = Util.Random.Next(20);
                enemy.LastPosition = enemy.Position = spawnPoint.Position;
                if (pp != null)
                    ((ZombieBear)enemy).SetSuper(1.5f);
                spawnedEnemies++;
                GameStatistics.CommonStats.ZombieTotals++;
            }
        }


        public void DespawnEnemy(Enemy enemy)
        {
            enemy.IsInUse = false;
            ((LevelTileNode)enemy.Parent).RemoveDynamicNode(enemy);
            enemy.Parent = null;
            spawnedEnemies--;
            enemy.Reset();

        }

        public float GetMovementSpeed(int difficulty)
        {
            return Util.Random.Next(MIN_SPEED[difficulty - 1], MAX_SPEED[difficulty - 1]);
        }

    }

}
